// Sudokus con Clase
// Resolución de sudokus con técnicas reales
// (C) 2022 Salvador Pozo
// Con Clase https://conclase.net
//
// Definición de la clase Tablero

#include "tablero.h"

Tablero::Tablero() : cuenta(-1) {
    casilla.resize(9);
    for(int f=0; f<9; f++) casilla[f].resize(9);
}

void Tablero::Leer(const unsigned char* t) {
    int f, c;

    cuenta = -1;
    finales = 0;
    puntero.Vaciar();
    for(f=0; f < 9; f++)
        for(c=0; c < 9; c++) {
            casilla[f][c] = Casilla(f, c, t[f*9+c]-'0', t[f*9+c] != '0');
            if(t[f*9+c] != '0') finales++;
        }
    for(f=0; f < 9; f++) fila[f].Iniciar(*this, f);
    for(c=0; c < 9; c++) columna[c].Iniciar(*this, c);
    for(c=0; c < 9; c++) caja[c].Iniciar(*this, c);
    LimpiaResaltes();
}

bool Tablero::Valido(int f, int c, int v) {
    return !fila[f].Presente(v) && !columna[c].Presente(v) && !caja[(f/3)*3+(c/3)].Presente(v);
}

void Tablero::Soluciones() {
    // Aplicar cribas
    bool aplicarfuerza=false;
    int valor;

    while(!aplicarfuerza) {
        valor=1; // Empezar las cribas con 1
        while(valor<10 && !Criba(valor)) valor++;
        if(valor<10) break;
        BorraCriba();
        BuscaParesDesnudos(); // Actualiza la lista de pares desnudos
        if(ProcesarParesDesnudos()) break;
        else aplicarfuerza=true;
    }
    FuerzaBruta();

    // Restablecer sudoku a estado incial
    while(!msgs.empty()) msgs.pop();
    finales = 0;
    puntero.Vaciar();
    LimpiaResaltes();
    BorraCriba();
    for(int k=0; k<9; k++) caja[k].IniciaPunteros();
    for(int f=0; f < 9; f++) {
        for(int c=0; c < 9; c++) {
            if(!casilla[f][c].Inicial()) {
                casilla[f][c].Lapiz() = MarcasLapiz();
                casilla[f][c].Asignar(0);
            } else finales++;
        }
    }
}

bool Tablero::FuerzaBruta() {
    int f, c, n;

    for(f=0; f<9; f++) {
        for(c=0; c<9; c++) {
            if(casilla[f][c].Valor() == 0) {
                for(n=1; n<=9; n++) {
                    if(Valido(f, c, n)) {
                        casilla[f][c].Asignar(n);
                        if(cuenta < 2) {
                            if(FuerzaBruta()) return true;
                            else casilla[f][c].Asignar(0);
                        }
                    }
                }
                return false;
            }
        }
    }
    cuenta++;
    return false;
}

bool Tablero::Asignar(int f, int c, int v) {
    if(!casilla[f][c].Valor()) {
        casilla[f][c].Asignar(v);
        casilla[f][c].Resaltar();
        // Eliminar marcas de lapiz en fila, columna y caja:
        int k = (f/3)*3 + (c/3);
        for(int i=0; i<9; i++) {
            if(fila[f].Cas(i)->Lapiz().EliminaMarca(v)) {
                fila[f].Cas(i)->Resaltar();
                fila[f].Cas(i)->DestruyeParDesnudo(); // Si la casilla forma paret de un par desnudo, destruir par desnudo
            }
            if(columna[c].Cas(i)->Lapiz().EliminaMarca(v)) {
                columna[c].Cas(i)->Resaltar();
                columna[c].Cas(i)->DestruyeParDesnudo(); // Si la casilla forma paret de un par desnudo, destruir par desnudo
            }
            if(caja[k].Cas(i)->Lapiz().EliminaMarca(v)) {
                caja[k].Cas(i)->Resaltar();
                caja[k].Cas(i)->DestruyeParDesnudo(); // Si la casilla forma paret de un par desnudo, destruir par desnudo
            }
        }
        finales++;
        return true;
    }
    return false;
}

bool Tablero::Asignar(Casilla &cas, int v) {
    return Asignar(cas.Fila(), cas.Columna(), v);
}

void Tablero::LimpiaResaltes() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            casilla[f][c].Lapiz().ResetEliminadas();
            casilla[f][c].Resaltar(false);
            casilla[f][c].Amarillo(false);
            casilla[f][c].Naranja(false);
        }
    }
}

void Tablero::ResetCriba() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(!casilla[f][c].Valor()) casilla[f][c].MarcaPosible();
            else casilla[f][c].MarcaNoPosible();
        }
    }
}

void Tablero::BorraCriba() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            casilla[f][c].MarcaNoPosible();
        }
    }
}

bool Tablero::Criba(int v) {
    int ncaja=0;

    ResetCriba();
    // Aplicar la criba a los pares desnudos:
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].ParDesnudo()) {
                if((casilla[f][c].Lapiz().ValorMarca(1) == v || casilla[f][c].Lapiz().ValorMarca(2) == v)) {
                    if(casilla[f][c].ParDesnudo(Casilla::enfila)) {
                        fila[f].Criba();
                        casilla[f][c].ParDesnudo(Casilla::enfila)->MarcaPosible();
                    }
                    if(casilla[f][c].ParDesnudo(Casilla::encolumna)) {
                        columna[c].Criba();
                        casilla[f][c].ParDesnudo(Casilla::encolumna)->MarcaPosible();
                    }
                    if(casilla[f][c].ParDesnudo(Casilla::encaja)) {
                        caja[casilla[f][c].Caja()].Criba();
                        casilla[f][c].ParDesnudo(Casilla::encaja)->MarcaPosible();
                    }
                    casilla[f][c].MarcaPosible();
                } else {
                    casilla[f][c].MarcaNoPosible();
                    if(casilla[f][c].ParDesnudo(Casilla::enfila)) casilla[f][c].ParDesnudo(Casilla::enfila)->MarcaNoPosible();
                    if(casilla[f][c].ParDesnudo(Casilla::encolumna)) casilla[f][c].ParDesnudo(Casilla::encolumna)->MarcaNoPosible();
                    if(casilla[f][c].ParDesnudo(Casilla::encaja)) casilla[f][c].ParDesnudo(Casilla::encaja)->MarcaNoPosible();
                }
            }
        }
    }
    // Aplicar criba a valores finales:
    for(int i=0; i<9; i++) {
        if(fila[i].Presente(v))    fila[i].Criba();
        if(columna[i].Presente(v)) columna[i].Criba();
        if(caja[i].Presente(v))    caja[i].Criba();
    }

    // Aplicar criba a parejas y trios punteros de cajas:
    // No desmarcar como posibles las casillas miembros de parejas y trios punteros:
    for(int f=0; f<3; f++) { // Fila de cajas
        for(int c=0; c<3; c++) { // Columna de cajas
            for(int i=0; i<3; i++) { // Fila de punteros
                if(caja[ncaja].PunteroFila(i, v)) {
                    fila[f*3+i].CribaMenosCaja(c);
                    for(int p=0; p<9; p++) { // Resto de la caja
                        if(p/3 != i) caja[ncaja].Cas(p)->MarcaNoPosible();
                    }
                }
            }
            for(int i=0; i<3; i++) { // Columna de punteros
                if(caja[ncaja].PunteroColumna(i, v)) {
                    columna[c*3+i].CribaMenosCaja(f);
                    for(int p=0; p<9; p++) { // Resto de la caja
                        if(p%3 != i) caja[ncaja].Cas(p)->MarcaNoPosible();
                    }
                }
            }
            ncaja++;
        }
    }
    return ProcesarCriba(v);
}

bool Tablero::ProcesarCriba(int v) {
    int f, c, k, p;
    int p1, p2, p3;
    int c1, c2, c3;
    int f1, f2, f3;

    for(k=0; k<9; k++) {
        switch(caja[k].CuentaPosibles()) {
        case 1:
            p=caja[k].UnicoPosible();
            Asignar(*caja[k].Cas(p), v);
            caja[k].Cas(p)->Resaltar();
            sprintf(m, "Valor final %d en %s" , v, caja[k].Cas(p)->Nombre());
            msgs.push(m);
            return true;
        case 2:
            p1 = caja[k].PrimeroPosible();
            p2 = caja[k].SegundoPosible(p1);
            if(!caja[k].Cas(p1)->Lapiz().ContieneMarca(v) || !caja[k].Cas(p2)->Lapiz().ContieneMarca(v)) {
                caja[k].Cas(p1)->Lapiz().AsignaMarca(v);
                caja[k].Cas(p2)->Lapiz().AsignaMarca(v);
                caja[k].Cas(p1)->Resaltar();
                caja[k].Cas(p2)->Resaltar();
                if(caja[k].Cas(p1)->Fila() == caja[k].Cas(p2)->Fila() && caja[k].AsignaPunteroFila(p1/3, v, 2))
                    sprintf(m, "Par puntero (%s, %s) %d", caja[k].Cas(p1)->Nombre(), caja[k].Cas(p2)->Nombre(), v);
                else if(caja[k].Cas(p1)->Columna() == caja[k].Cas(p2)->Columna() && caja[k].AsignaPunteroColumna(p1%3, v, 2))
                    sprintf(m, "Par puntero (%s, %s) %d", caja[k].Cas(p1)->Nombre(), caja[k].Cas(p2)->Nombre(), v);
                else sprintf(m, "Pareja de candidatos (%s, %s) %d", caja[k].Cas(p1)->Nombre(), caja[k].Cas(p2)->Nombre(), v);
                msgs.push(m);
                VerificarParDesnudo(caja[k].Cas(p1), caja[k].Cas(p2), v);
                return true;
            }
            break;
        case 3:
            p1 = caja[k].PrimeroPosible();
            p2 = caja[k].SegundoPosible(p1);
            p3 = caja[k].TerceroPosible(p2);
            // Si las tres casillas están en la misma fila o columna y no había sido detectado antes, añadir trio puntero
            if((caja[k].Cas(p1)->Fila() == caja[k].Cas(p2)->Fila() && caja[k].Cas(p1)->Fila() == caja[k].Cas(p3)->Fila() && caja[k].AsignaPunteroFila(p1/3, v, 3)) ||
               (caja[k].Cas(p1)->Columna() == caja[k].Cas(p2)->Columna() && caja[k].Cas(p1)->Columna() == caja[k].Cas(p3)->Columna() && caja[k].AsignaPunteroColumna(p1%3, v, 3))) {
                caja[k].Cas(p1)->Resaltar();
                caja[k].Cas(p2)->Resaltar();
                caja[k].Cas(p3)->Resaltar();
                sprintf(m, "Trio puntero (%s, %s, %s) %d", caja[k].Cas(p1)->Nombre(), caja[k].Cas(p2)->Nombre(), caja[k].Cas(p3)->Nombre(), v);
                msgs.push(m);
                return true;
            }
            break;
        }
    }
    for(f=0; f<9; f++) {
        // Contar posibles en cada fila
        switch(fila[f].CuentaPosibles()) {
        case 1: // Valor final
            c=fila[f].UnicoPosible();
            Asignar(f, c, v);
            casilla[f][c].Resaltar();
            sprintf(m, "Valor final %d en %s" , v, casilla[f][c].Nombre());
            msgs.push(m);
            return true;
        case 2: // Posible par puntero
            c1 = fila[f].PrimeroPosible();
            c2 = fila[f].SegundoPosible(c1);
            k = casilla[f][c1].Caja();
            // Si las dos casillas están en la misma caja y no había sido detectado antes, añadir par puntero
            if(k==casilla[f][c2].Caja() && caja[k].AsignaPunteroFila(f%3, v, 2)) {
                casilla[f][c1].Lapiz().AsignaMarca(v);
                casilla[f][c2].Lapiz().AsignaMarca(v);
                casilla[f][c1].Resaltar();
                casilla[f][c2].Resaltar();
                sprintf(m, "Par puntero (%s, %s) %d", casilla[f][c1].Nombre(), casilla[f][c2].Nombre(), v);
                msgs.push(m);
                // Verificar además si se trata de un par desnudo:
                VerificarParDesnudo(&casilla[f][c1], &casilla[f][c2], v);
                return true;
            }
            break;
        case 3: // Posible trio puntero
            c1 = fila[f].PrimeroPosible();
            c2 = fila[f].SegundoPosible(c1);
            c3 = fila[f].TerceroPosible(c2);
            k = casilla[f][c1].Caja();
            // Si las tres casillas están en la misma caja y no había sido detectado antes, añadir trio puntero
            if(k==casilla[f][c2].Caja() && k==casilla[f][c3].Caja() && caja[k].AsignaPunteroFila(f%3, v, 3)) {
                casilla[f][c1].Resaltar();
                casilla[f][c2].Resaltar();
                casilla[f][c3].Resaltar();
                sprintf(m, "Trio puntero (%s, %s, %s) %d", casilla[f][c1].Nombre(), casilla[f][c2].Nombre(), casilla[f][c3].Nombre(), v);
                msgs.push(m);
                return true;
            }
            break;
        }
    }
    for(c=0; c<9; c++) {
        // Contar posibles en cada columna
        switch(columna[c].CuentaPosibles()) {
        case 1: // Valor final
            f=columna[c].UnicoPosible();
            Asignar(f, c, v);
            casilla[f][c].Resaltar();
            sprintf(m, "Valor final %d en %s" , v, casilla[f][c].Nombre());
            msgs.push(m);
            return true;
        case 2: // Posible par puntero
            f1 = columna[c].PrimeroPosible();
            f2 = columna[c].SegundoPosible(f1);
            k = casilla[f1][c].Caja();
            // Si las dos casillas están en la misma caja y no había sido detectado antes, añadir par puntero
            if(k==casilla[f2][c].Caja() && caja[k].AsignaPunteroColumna(c%3, v, 2)) {
                casilla[f1][c].Lapiz().AsignaMarca(v);
                casilla[f2][c].Lapiz().AsignaMarca(v);
                casilla[f1][c].Resaltar();
                casilla[f2][c].Resaltar();
                sprintf(m, "Par puntero (%s, %s) %d", casilla[f1][c].Nombre(), casilla[f2][c].Nombre(), v);
                msgs.push(m);
                // Verificar además si se trata de un par desnudo:
                VerificarParDesnudo(&casilla[f1][c], &casilla[f2][c], v);
                return true;
            }
            break;
        case 3: // Posible trio puntero
            f1 = columna[c].PrimeroPosible();
            f2 = columna[c].SegundoPosible(f1);
            f3 = columna[c].TerceroPosible(f2);
            k = casilla[f1][c].Caja();
            // Si las tres casillas están en la misma caja y no había sido detectado antes, añadir trio puntero
            if(k==casilla[f2][c].Caja() && k==casilla[f3][c].Caja() && caja[k].AsignaPunteroColumna(c%3, v, 3)) {
                casilla[f1][c].Resaltar();
                casilla[f2][c].Resaltar();
                casilla[f3][c].Resaltar();
                sprintf(m, "Trio puntero (%s, %s, %s) %d", casilla[f1][c].Nombre(), casilla[f2][c].Nombre(), casilla[f3][c].Nombre(), v);
                msgs.push(m);
                return true;
            }
            break;
        }
    }
    return false;
}

void Tablero::VerificarParDesnudo(Casilla* c1, Casilla* c2, int v) {
    int v2;

    for(v2=1; v2<=9; v2++) {
        if(v2!=v && c1->Lapiz().ContieneMarca(v2) && c2->Lapiz().ContieneMarca(v2)) {
            if(c1->Fila() == c2->Fila()) c1->MarcaParDesnudo(c2, Casilla::enfila);
            else if(c1->Columna() == c2->Columna()) c1->MarcaParDesnudo(c2, Casilla::encolumna);
            c1->MarcaParDesnudo(c2, Casilla::encaja); // En estos casos, el par desnudo SIEMPRE estará en una caja
            sprintf(m, "Encontrado par desnudo en (%s,%s) valores [%d,%d]", c1->Nombre(), c2->Nombre(), v2, v);
            msgs.push(m);
            // Eliminar resto de marcas del par desnudo
            for(int v3=1; v3<=9; v3++) {
                if(v3!=v && v3!=v2) {
                    if(c1->Lapiz().EliminaMarca(v3)) {
                        sprintf(m, "Eliminar candidato %d en %s", v3, c1->Nombre());
                        msgs.push(m);

                    }
                    if(c2->Lapiz().EliminaMarca(v3)) {
                        sprintf(m, "Eliminar candidato %d en %s", v3, c2->Nombre());
                        msgs.push(m);

                    }
                }
            }
            break; // Optimizar
        }
    }
}

bool Tablero::BuscaParesDesnudos(bool todos) {
    bool retval=false;

    if(todos) {
        retval |= BuscaParesDesnudosFilas();
        retval |= BuscaParesDesnudosColumnas();
    }
    retval |= BuscaParesDesnudosCajas();

    return retval;
}

bool Tablero::BuscaParesDesnudosFilas() {
    bool retval=false;

    for(int f=0; f<9; f++) {
        for(int c=0; c<8; c++) {
            // Solo buscamos en casillas vacías con dos marcas de lápiz.
            if(casilla[f][c].Lapiz().CuentaMarcas() == 2) {
                for(int c2=c+1; c2<9; c2++) {
                    // Buscamos en el resto de casillas vacías con dos valores posibles
                    if(casilla[f][c2].Lapiz() == casilla[f][c].Lapiz() && !casilla[f][c].ParDesnudo(Casilla::enfila)) {
                        casilla[f][c].MarcaParDesnudo(&casilla[f][c2], Casilla::enfila);
                        if(c/3==c2/3) casilla[f][c].MarcaParDesnudo(&casilla[f][c2], Casilla::encaja);
                        sprintf(m, "Encontrado par desnudo en (%s,%s) valores [%s]", casilla[f][c].Nombre(), casilla[f][c2].Nombre(), casilla[f][c].Lapiz().Valores());
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaParesDesnudosColumnas() {
    bool retval = false;

    for(int c=0; c<9; c++) {
        for(int f=0; f<8; f++) {
            if(casilla[f][c].Lapiz().CuentaMarcas() == 2) {
                for(int f2=f+1; f2<9; f2++) {
                    // Buscamos en el resto de casillas vacías con dos valores posibles
                    if(casilla[f2][c].Lapiz() == casilla[f][c].Lapiz() && !casilla[f][c].ParDesnudo(Casilla::encolumna)) {
                        casilla[f][c].MarcaParDesnudo(&casilla[f2][c], Casilla::encolumna);
                        if(f/3==f2/3) casilla[f][c].MarcaParDesnudo(&casilla[f2][c], Casilla::encaja);
                        sprintf(m, "Encontrado par desnudo en (%s,%s) valores [%s]", casilla[f][c].Nombre(), casilla[f2][c].Nombre(), casilla[f][c].Lapiz().Valores());
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaParesDesnudosCajas() {
    bool retval = false;

    for(int k=0; k<9; k++) {
        for(int p=0; p<8; p++) {
            if(caja[k].Cas(p)->Lapiz().CuentaMarcas() == 2) {
                for(int p2=p+1; p2<9; p2++) {
                    // Buscamos en el resto de casillas vacías con dos valores posibles
                    if(caja[k].Cas(p2)->Lapiz() == caja[k].Cas(p)->Lapiz() && !caja[k].Cas(p)->ParDesnudo(Casilla::encaja)) {
                        caja[k].Cas(p)->MarcaParDesnudo(caja[k].Cas(p2), Casilla::encaja);
                        //caja[k].Cas(p2)->MarcaParDesnudo(caja[k].Cas(p));
                        sprintf(m, "Encontrado par desnudo en (%s,%s) valores [%s]", caja[k].Cas(p)->Nombre(), caja[k].Cas(p2)->Nombre(), caja[k].Cas(p)->Lapiz().Valores());
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesarParesDesnudos(bool todos) {
    for(int f=0; f<9; f++) if(ProcesarDesnudoFila(f, todos)) return true;
    for(int c=0; c<9; c++) if(ProcesarDesnudoColumna(c, todos)) return true;
    for(int k=0; k<9; k++) if(ProcesarDesnudoCaja(k)) return true;
    return false;
}

bool Tablero::ProcesarDesnudoFila(int f, bool todos) {
    bool presente[9] = { false, false, false, false, false, false, false, false, false };
    bool ocupada[9] = { false, false, false, false, false, false, false, false, false };
    int libre[2];
    int valor[2];
    int ocupadas=0;

    for(int c=0; c<9; c++) {
        if(casilla[f][c].Valor()) {
            ocupadas++;
            presente[casilla[f][c].Valor()-1] = true;
            ocupada[c] = true;
        }
        if(casilla[f][c].ParDesnudo(Casilla::enfila)) {
            ocupadas++;
            presente[casilla[f][c].Lapiz().ValorMarca(1)-1] = true;
            presente[casilla[f][c].Lapiz().ValorMarca(2)-1] = true;
            ocupada[c] = true;
        }
    }

    if(ocupadas<7 || ocupadas==9) return false;
    for(libre[0]=0; ocupada[libre[0]]; libre[0]++); // Localiza la casilla libre
    for(valor[0]=0; presente[valor[0]]; valor[0]++);  // Localiza el valor ausente

    if(ocupadas==8) { // Valor final:
        fila[f].Cas(libre[0])->Resaltar();
        Asignar(f, libre[0], valor[0]+1);
        sprintf(m, "Valor final %d en %s", valor[0]+1, fila[f].Cas(libre[0])->Nombre());
        msgs.push(m);
        return true;
    }

    // Otro par desnudo
    for(libre[1]=libre[0]+1; ocupada[libre[1]]; libre[1]++); // Localiza la segunda casilla libre
    if(todos || libre[0]/3 == libre[1]/3) { // En primera fase sólo almacenamos pares desnudos en la misma caja
        // Amarillear pares desnudos de fila
        for(int c=0; c<9; c++) {
            if(casilla[f][c].ParDesnudo(Casilla::enfila)) {
                casilla[f][c].Amarillo();
            }
        }
        for(valor[1]=valor[0]+1; presente[valor[1]]; valor[1]++);  // Localiza el segundo valor ausente
        fila[f].Cas(libre[0])->Resaltar();
        fila[f].Cas(libre[1])->Resaltar();

        fila[f].Cas(libre[0])->MarcaParDesnudo(fila[f].Cas(libre[1]), Casilla::enfila);
        if(libre[0]/3 == libre[1]/3) fila[f].Cas(libre[0])->MarcaParDesnudo(fila[f].Cas(libre[1]), Casilla::encaja);

        fila[f].Cas(libre[0])->Lapiz().AsignaMarca(valor[0]+1);
        fila[f].Cas(libre[0])->Lapiz().AsignaMarca(valor[1]+1);
        fila[f].Cas(libre[1])->Lapiz().AsignaMarca(valor[0]+1);
        fila[f].Cas(libre[1])->Lapiz().AsignaMarca(valor[1]+1);
        sprintf(m, "Par desnudo en en %s,%s, %d,%d", fila[f].Cas(libre[0])->Nombre(), fila[f].Cas(libre[1])->Nombre(), valor[0]+1, valor[1]+1);
        msgs.push(m);

        // Eliminar resto de marcas del par desnudo
        for(int v3=1; v3<=9; v3++) {
            if(v3!=valor[0]+1 && v3!=valor[1]+1) {
                if(fila[f].Cas(libre[0])->Lapiz().EliminaMarca(v3)) {
                    sprintf(m, "Eliminar candidato %d en %s", v3, fila[f].Cas(libre[0])->Nombre());
                    msgs.push(m);

                }
                if(fila[f].Cas(libre[1])->Lapiz().EliminaMarca(v3)) {
                    sprintf(m, "Eliminar candidato %d en %s", v3, fila[f].Cas(libre[0])->Nombre());
                    msgs.push(m);

                }
            }
        }

        return true;
    }
    return false;
}

bool Tablero::ProcesarDesnudoColumna(int c, bool todos) {
    bool presente[9] = { false, false, false, false, false, false, false, false, false };
    bool ocupada[9] = { false, false, false, false, false, false, false, false, false };
    int libre[2];
    int valor[2];
    int ocupadas=0;

    for(int f=0; f<9; f++) {
        if(casilla[f][c].Valor()) {
            ocupadas++;
            presente[casilla[f][c].Valor()-1] = true;
            ocupada[f] = true;
        }
        if(casilla[f][c].ParDesnudo(Casilla::encolumna)) {
            ocupadas++;
            presente[casilla[f][c].Lapiz().ValorMarca(1)-1] = true;
            presente[casilla[f][c].Lapiz().ValorMarca(2)-1] = true;
            ocupada[f] = true;
        }
    }

    if(ocupadas<7 || ocupadas==9) return false;
    for(libre[0]=0; ocupada[libre[0]]; libre[0]++); // Localiza la casilla libre
    for(valor[0]=0; presente[valor[0]]; valor[0]++);  // Localiza el valor ausente

    if(ocupadas==8) { // Valor final:
        columna[c].Cas(libre[0])->Resaltar();
        Asignar(libre[0], c, valor[0]+1);
        sprintf(m, "Valor final %d en %s", valor[0]+1, columna[c].Cas(libre[0])->Nombre());
        msgs.push(m);
        return true;
    }

    // Otro par desnudo
    for(libre[1]=libre[0]+1; ocupada[libre[1]]; libre[1]++); // Localiza la segunda casilla libre
    if(todos || libre[0]/3 == libre[1]/3) { // En primera fase sólo almacenamos pares desnudos en la misma caja
        // Amarillear pares desnudos de fila
        for(int f=0; f<9; f++) {
            if(casilla[f][c].ParDesnudo(Casilla::encolumna)) {
                casilla[f][c].Amarillo();
            }
        }
        for(valor[1]=valor[0]+1; presente[valor[1]]; valor[1]++);  // Localiza el segundo valor ausente
        columna[c].Cas(libre[0])->Resaltar();
        columna[c].Cas(libre[1])->Resaltar();
        columna[c].Cas(libre[0])->MarcaParDesnudo(columna[c].Cas(libre[1]), Casilla::encolumna);
        if(libre[0]/3 == libre[1]/3) columna[c].Cas(libre[0])->MarcaParDesnudo(columna[c].Cas(libre[1]), Casilla::encaja);
        columna[c].Cas(libre[0])->Lapiz().AsignaMarca(valor[0]+1);
        columna[c].Cas(libre[0])->Lapiz().AsignaMarca(valor[1]+1);
        columna[c].Cas(libre[1])->Lapiz().AsignaMarca(valor[0]+1);
        columna[c].Cas(libre[1])->Lapiz().AsignaMarca(valor[1]+1);
        sprintf(m, "Par desnudo en en %s,%s, %d,%d", columna[c].Cas(libre[0])->Nombre(), columna[c].Cas(libre[1])->Nombre(), valor[0]+1, valor[1]+1);
        msgs.push(m);

        // Eliminar resto de marcas del par desnudo
        for(int v3=1; v3<=9; v3++) {
            if(v3!=valor[0]+1 && v3!=valor[1]+1) {
                if(columna[c].Cas(libre[0])->Lapiz().EliminaMarca(v3)) {
                    sprintf(m, "Eliminar candidato %d en %s", v3, columna[c].Cas(libre[0])->Nombre());
                    msgs.push(m);

                }
                if(columna[c].Cas(libre[1])->Lapiz().EliminaMarca(v3)) {
                    sprintf(m, "Eliminar candidato %d en %s", v3, columna[c].Cas(libre[0])->Nombre());
                    msgs.push(m);

                }
            }
        }

        return true;
    }
    return false;
}

bool Tablero::ProcesarDesnudoCaja(int k) {
    bool presente[9] = { false, false, false, false, false, false, false, false, false };
    bool ocupada[9] = { false, false, false, false, false, false, false, false, false };
    int libre[2];
    int valor[2];
    int ocupadas=0;

    for(int p=0; p<9; p++) {
        if(caja[k].Cas(p)->Valor()) {
            ocupadas++;
            presente[caja[k].Cas(p)->Valor()-1] = true;
            ocupada[p] = true;
        }
        if(caja[k].Cas(p)->ParDesnudo(Casilla::encaja)) {
            ocupadas++;
            presente[caja[k].Cas(p)->Lapiz().ValorMarca(1)-1] = true;
            presente[caja[k].Cas(p)->Lapiz().ValorMarca(2)-1] = true;
            ocupada[p] = true;
        }
    }

    if(ocupadas<7 || ocupadas==9) return false;
    for(libre[0]=0; ocupada[libre[0]]; libre[0]++); // Localiza la casilla libre
    for(valor[0]=0; presente[valor[0]]; valor[0]++);  // Localiza el valor ausente
    if(ocupadas==8) { // Valor final:
        caja[k].Cas(libre[0])->Resaltar();
        Asignar(*caja[k].Cas(libre[0]), valor[0]+1);
        sprintf(m, "Valor final %d en %s", valor[0]+1, caja[k].Cas(libre[0])->Nombre());
        msgs.push(m);
        return true;
    }
    // Otro par desnudo
    for(libre[1]=libre[0]+1; ocupada[libre[1]]; libre[1]++); // Localiza la segunda casilla libre
    for(valor[1]=valor[0]+1; presente[valor[1]]; valor[1]++);  // Localiza el segundo valor ausente
    caja[k].Cas(libre[0])->Resaltar();
    caja[k].Cas(libre[1])->Resaltar();
    caja[k].Cas(libre[0])->MarcaParDesnudo(caja[k].Cas(libre[1]), Casilla::encaja);
    //caja[k].Cas(libre[1])->MarcaParDesnudo(caja[k].Cas(libre[0]));
    caja[k].Cas(libre[0])->Lapiz().AsignaMarca(valor[0]+1);
    caja[k].Cas(libre[0])->Lapiz().AsignaMarca(valor[1]+1);
    caja[k].Cas(libre[1])->Lapiz().AsignaMarca(valor[0]+1);
    caja[k].Cas(libre[1])->Lapiz().AsignaMarca(valor[1]+1);
    sprintf(m, "Par desnudo en en %s,%s, %d,%d", caja[k].Cas(libre[0])->Nombre(), caja[k].Cas(libre[1])->Nombre(), valor[0]+1, valor[1]+1);
    msgs.push(m);
    // Eliminar resto de marcas del par desnudo
    for(int v3=1; v3<=9; v3++) {
        if(v3!=valor[0]+1 && v3!=valor[1]+1) {
            if(caja[k].Cas(libre[0])->Lapiz().EliminaMarca(v3)) {
                sprintf(m, "Eliminar candidato %d en %s", v3, caja[k].Cas(libre[0])->Nombre());
                msgs.push(m);

            }
            if(caja[k].Cas(libre[1])->Lapiz().EliminaMarca(v3)) {
                sprintf(m, "Eliminar candidato %d en %s", v3, caja[k].Cas(libre[0])->Nombre());
                msgs.push(m);

            }
        }
    }
    return true;
}

// Procedimientos de fase 2:
void Tablero::PonerMarcasLapiz() {
    sprintf(m, "Poner todas las marcas de lapiz");
    msgs.push(m);
    int v1, v2;
    int c1, c2;
    int f1, f2;
    int k, p1, p2;

    // Para cada casilla colocar marcas de lápiz que faltan:
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(!casilla[f][c].Valor()) { // Si no tiene un valor asignado
                for(int v=1; v<=9; v++) {
                    int k = casilla[f][c].Caja();
                    if(!fila[f].Presente(v) && !columna[c].Presente(v) && !caja[k].Presente(v)) {
                        // No añadir marcas a casillas que formen parte de un par desnudo
                        if(!casilla[f][c].ParDesnudo())
                            casilla[f][c].Lapiz().AsignaMarca(v);
                    }
                }
            } else {
                casilla[f][c].Lapiz() = MarcasLapiz();
            }
        }
    }

    // Procesar punteros:
    for(int v=1; v<=9; v++) {
        int ncaja=0;
        for(int f=0; f<3; f++) { // Fila de cajas
            for(int c=0; c<3; c++) { // Columna de cajas
                for(int i=0; i<3; i++) { // Fila de punteros
                    if(caja[ncaja].PunteroFila(i, v)) {
                        fila[f*3+i].DesmarcaLapizMenosCaja(c, v);
                        for(int p=0; p<9; p++) { // Resto de la caja
                            if(p/3 != i) caja[ncaja].Cas(p)->Lapiz().EliminaMarca(v);
                        }
                    }
                }
                for(int i=0; i<3; i++) { // Columna de punteros
                    if(caja[ncaja].PunteroColumna(i, v)) {
                        columna[c*3+i].DesmarcaLapizMenosCaja(f, v);
                        for(int p=0; p<9; p++) { // Resto de la caja
                            if(p%3 != i) caja[ncaja].Cas(p)->Lapiz().EliminaMarca(v);
                        }
                    }
                }
                ncaja++;
            }
        }
    }

    // Aplicar la criba a los pares desnudos:
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].ParDesnudo()) {
                v1 = casilla[f][c].Lapiz().ValorMarca(1);
                v2 = casilla[f][c].Lapiz().ValorMarca(2);
                if(casilla[f][c].ParDesnudo(Casilla::enfila)) { // Eliminar v1 y v2 como candidatos en resto de fila
                    c1 = c;
                    c2 = casilla[f][c].ParDesnudo(Casilla::enfila)->Columna();
                    for(int c3=0; c3<9; c3++) {
                        if(c3 != c1 && c3 != c2) {
                            fila[f].Cas(c3)->Lapiz().EliminaMarca(v1);
                            fila[f].Cas(c3)->Lapiz().EliminaMarca(v2);
                        }
                    }
                }
                if(casilla[f][c].ParDesnudo(Casilla::encolumna)) { // Eliminar v1 y v2 como candidatos en resto de columna
                    f1 = f;
                    f2 = casilla[f][c].ParDesnudo(Casilla::encolumna)->Fila();
                    for(int f3=0; f3<9; f3++) {
                        if(f3 != f1 && f3 != f2) {
                            columna[c].Cas(f3)->Lapiz().EliminaMarca(v1);
                            columna[c].Cas(f3)->Lapiz().EliminaMarca(v2);
                        }
                    }
                }
                if(casilla[f][c].ParDesnudo(Casilla::encaja)) {  // Eliminar v1 y v2 como candidatos en resto de caja
                    k = casilla[f][c].Caja();
                    p1 = casilla[f][c].PosEnCaja();
                    p2 = casilla[f][c].ParDesnudo(Casilla::encaja)->PosEnCaja();
                    for(int p3=0; p3<9; p3++) {
                        if(p3 != p1 && p3 != p2) {
                            caja[k].Cas(p3)->Lapiz().EliminaMarca(v1);
                            caja[k].Cas(p3)->Lapiz().EliminaMarca(v2);
                        }
                    }
                }
            }
        }
    }
}

bool Tablero::CasillasConUnaMarcaLapiz() {
    int v;

    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].Lapiz().CuentaMarcas() == 1) {
                v=casilla[f][c].Lapiz().Valor();
                Asignar(f, c, v);
                sprintf(m, "Unica Marca Lapiz %c%c, valor %d", 'A'+f, '1'+c, v);
                msgs.push(m);
                return true;
            }
        }
    }
    return false;
}

bool Tablero::UnicaCasillaConCandidato() {
    Unico u(0,0);

    // Buscar casillas con único valor de candidato en filas, columnas y cajas:
    for(int f=0; f<9; f++) {
        u = fila[f].UnicaCasillaConCandidato();
        if(u.Casilla() != -1) {
            Asignar(f, u.Casilla(), u.Valor());
            sprintf(m, "Unica Casilla Candidato Fila %c%c, valor %d", 'A'+f, '1'+u.Casilla(), u.Valor());
            msgs.push(m);
            return true;
        }
    }
    for(int c=0; c<9; c++) {
        u = columna[c].UnicaCasillaConCandidato();
        if(u.Casilla() != -1) {
            Asignar(u.Casilla(), c, u.Valor());
            sprintf(m, "Unica Casilla Candidato Columna %c%c, valor %d", 'A'+u.Casilla(), '1'+c, u.Valor());
            msgs.push(m);
            return true;
        }
    }
    for(int k=0; k<9; k++) {
        u = caja[k].UnicaCasillaConCandidato();
        if(u.Casilla() != -1) {
            int fil = caja[k].Cas(u.Casilla())->Fila();
            int col = caja[k].Cas(u.Casilla())->Columna();
            Asignar(fil, col, u.Valor());
            sprintf(m, "Unica Casilla Candidato Caja %d %c%c, valor %d", k+1, 'A'+fil, '1'+col, u.Valor());
            msgs.push(m);
            return true;
        }
    }
    return false;
}

// Buscar pares y trios puntero
bool Tablero::BuscaPunteros() {
    bool retval = false;

    for(int v=1; v<=9; v++) {
        for(int f=0; f<9; f++)
            retval |= BuscaPunterosFila(f, v);
        for(int c=0; c<9; c++)
            retval |= BuscaPunterosColumna(c, v);
        for(int k=0; k<9; k++)
            retval |= BuscaPunterosCaja(k, v);
    }
    return retval;
}

// Busca el candidato v en la fila f, si está dos o tres veces en la misma caja lo guarda.
bool Tablero::BuscaPunterosFila(int f, int v) {
    int cuenta = 0;

    // Si el valor v está asignado a la fila, retornar false
    if(!fila[f].Presente(v)) {
        // Leer casillas con la marca de lápiz v. Si son más de tres retornar false
        for(int c=0; c<9; c++) if(fila[f].Cas(c)->Lapiz().ContieneMarca(v)) cuenta++;
        if(cuenta == 3 || cuenta == 2) {
            Puntero pun(cuenta);
            cuenta = 0;
            pun.AsignarVal(v);
            for(int c=0; c<9; c++) {
                if(fila[f].Cas(c)->Lapiz().ContieneMarca(v)) {
                    pun.AsignarCas(cuenta++, fila[f].Cas(c));
                }
            }
            // Si todas las casillas están en la misma caja, almacenar.
            if(pun.MismaCaja()!=-1) {
                puntero.Insertar(pun);
                return true;
            }
        }
    }
    return false;
}

bool Tablero::BuscaPunterosColumna(int c, int v) {
    int cuenta = 0;

    // Si el valor v está asignado a la fila, retornar false
    if(!columna[c].Presente(v)) {
        // Leer casillas con la marca de lápiz v. Si son más de tres retornar false
        for(int f=0; f<9; f++) if(columna[c].Cas(f)->Lapiz().ContieneMarca(v)) cuenta++;
        if(cuenta == 3 || cuenta == 2) {
            Puntero pun(cuenta);
            cuenta = 0;
            pun.AsignarVal(v);
            for(int f=0; f<9; f++) {
                if(columna[c].Cas(f)->Lapiz().ContieneMarca(v)) {
                    pun.AsignarCas(cuenta++, columna[c].Cas(f));
                }
            }
            // Si todas las casillas están en la misma caja, almacenar.
            if(pun.MismaCaja()!=-1) {
                puntero.Insertar(pun);
                return true;
            }
        }
    }
    return false;
}

bool Tablero::BuscaPunterosCaja(int k, int v) {
    int cuenta = 0;

    // Si el valor v está asignado a la fila, retornar false
    if(!caja[k].Presente(v)) {
        // Leer casillas con la marca de lápiz v. Si son más de tres retornar false
        for(int p=0; p<9; p++) if(caja[k].Cas(p)->Lapiz().ContieneMarca(v)) cuenta++;
        if(cuenta == 3 || cuenta == 2) {
            Puntero pun(cuenta);
            cuenta = 0;
            pun.AsignarVal(v);
            for(int p=0; p<9; p++) {
                if(caja[k].Cas(p)->Lapiz().ContieneMarca(v)) {
                    pun.AsignarCas(cuenta++, caja[k].Cas(p));
                }
            }
            // Si todas las casillas están en la misma caja, almacenar.
            if(pun.MismaFila()!=-1 || pun.MismaColumna()!=-1) {
                puntero.Insertar(pun);
                return true;
            }
        }
    }
    return false;
}

bool Tablero::ProcesaPunterosFila(int f) {
    bool retval=false;

    // Procesar punteros
    ListaPuntero()->Primero();
    while(ListaPuntero()->Actual()) {
        int fil = ListaPuntero()->ValorActual().MismaFila();
        if(fil == f) {
            for(int c=0; c<9; c++) { // Comparar casillas?? casilla[f][c] != ListaPuntero()->ValorActual().Cas(0)
                if(c != ListaPuntero()->ValorActual().Cas(0)->Columna() &&
                   c != ListaPuntero()->ValorActual().Cas(1)->Columna() &&
                   (!ListaPuntero()->ValorActual().Trio() || c!= ListaPuntero()->ValorActual().Cas(2)->Columna())) {
                    if(casilla[f][c].Lapiz().EliminaMarca(ListaPuntero()->ValorActual().Val())) {
                        if(!retval) {
                            if(puntero.ValorActual().Par())
                                sprintf(m, "Par puntero valor %d en (%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre());
                            else
                                sprintf(m, "Trio puntero valor %d en (%s,%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre(), puntero.ValorActual().Cas(2)->Nombre());

                            msgs.push(m);
                        }
                        casilla[f][c].Resaltar();
                        ListaPuntero()->ValorActual().Cas(0)->Amarillo();
                        ListaPuntero()->ValorActual().Cas(1)->Amarillo();
                        if(ListaPuntero()->ValorActual().Trio()) ListaPuntero()->ValorActual().Cas(2)->Amarillo();
                        sprintf(m, "Eliminar candidato %d en %c%c", ListaPuntero()->ValorActual().Val(), 'A'+f, '1'+c);
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
            if(retval) return true;
        }
        ListaPuntero()->Siguiente();
    }
    return false;
}

bool Tablero::ProcesaPunterosColumna(int c) {
    bool retval=false;

    // Procesar punteros
    ListaPuntero()->Primero();
    while(ListaPuntero()->Actual()) {
        int col = ListaPuntero()->ValorActual().MismaColumna();
        if(col == c) {
            for(int f=0; f<9; f++) {
                if(f != ListaPuntero()->ValorActual().Cas(0)->Fila() &&
                   f != ListaPuntero()->ValorActual().Cas(1)->Fila()&&
                   (!ListaPuntero()->ValorActual().Trio() || f!= ListaPuntero()->ValorActual().Cas(2)->Fila())) {
                    if(casilla[f][c].Lapiz().EliminaMarca(ListaPuntero()->ValorActual().Val())) {
                        if(!retval) {
                            if(puntero.ValorActual().Par())
                                sprintf(m, "Par puntero valor %d en (%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre());
                            else
                                sprintf(m, "Trio puntero valor %d en (%s,%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre(), puntero.ValorActual().Cas(2)->Nombre());

                            msgs.push(m);
                        }
                        casilla[f][c].Resaltar();
                        ListaPuntero()->ValorActual().Cas(0)->Amarillo();
                        ListaPuntero()->ValorActual().Cas(1)->Amarillo();
                        if(ListaPuntero()->ValorActual().Trio()) ListaPuntero()->ValorActual().Cas(2)->Amarillo();
                        sprintf(m, "Eliminar candidato %d en %c%c", ListaPuntero()->ValorActual().Val(), 'A'+f, '1'+c);
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
            if(retval) return true;
        }
        ListaPuntero()->Siguiente();
    }
    return false;
}

bool Tablero::ProcesaPunterosCaja(int k) {
    bool retval = false;
    // Procesar punteros
    ListaPuntero()->Primero();
    while(ListaPuntero()->Actual()) {
        int caj = ListaPuntero()->ValorActual().MismaCaja();
        if(caj == k) {
            for(int p=0; p<9; p++) {
                if(p != ListaPuntero()->ValorActual().Cas(0)->PosEnCaja() &&
                   p != ListaPuntero()->ValorActual().Cas(1)->PosEnCaja() &&
                   (!ListaPuntero()->ValorActual().Trio() || p!= ListaPuntero()->ValorActual().Cas(2)->PosEnCaja())) {
                    if(caja[k].Cas(p)->Lapiz().EliminaMarca(ListaPuntero()->ValorActual().Val())) {
                        if(!retval) {
                            if(puntero.ValorActual().Par())
                                sprintf(m, "Par puntero valor %d en (%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre());
                            else
                                sprintf(m, "Trio puntero valor %d en (%s,%s,%s)", puntero.ValorActual().Val(),
                                    puntero.ValorActual().Cas(0)->Nombre(), puntero.ValorActual().Cas(1)->Nombre(), puntero.ValorActual().Cas(2)->Nombre());

                            msgs.push(m);
                        }
                        int f = caja[k].Cas(p)->Fila();
                        int c = caja[k].Cas(p)->Columna();
                        casilla[f][c].Resaltar();
                        ListaPuntero()->ValorActual().Cas(0)->Amarillo();
                        ListaPuntero()->ValorActual().Cas(1)->Amarillo();
                        if(ListaPuntero()->ValorActual().Trio()) ListaPuntero()->ValorActual().Cas(2)->Amarillo();
                        sprintf(m, "Eliminar candidato %d en %c%c", ListaPuntero()->ValorActual().Val(), 'A'+f, '1'+c);
                        msgs.push(m);
                        retval = true;
                    }
                }
            }
            if(retval) return true;
        }
        ListaPuntero()->Siguiente();
    }
    return false;
}

bool Tablero::ProcesaPunteros() {
    for(int f=0; f<9; f++) if(ProcesaPunterosFila(f)) return true;
    for(int c=0; c<9; c++) if(ProcesaPunterosColumna(c)) return true;
    for(int k=0; k<9; k++) if(ProcesaPunterosCaja(k)) return true;
    return false;
}

bool Tablero::ProcesaGrupoFila(int f, int nElementos) {
    Oculto grupo(fila[f].GetCasillas());
    MarcasLapiz marcas;
    bool retval = false;

    if(grupo.PrimeraCombinacion(nElementos)) {
        do {
            marcas = MarcasLapiz();
            for(int i = 0; i<nElementos; i++) {
                if(grupo.Comb(i)->Valor()) marcas.MarcarTodas(); // Ignorar grupos con casillas finales
                else marcas = marcas | grupo.Comb(i)->Lapiz();
            }
            if(marcas.CuentaMarcas() == nElementos) {
                for(int i = 0; i<9; i++) {
                    if(!grupo.Contiene(fila[f].Cas(i), nElementos)) {
                        retval |= (fila[f].Cas(i)->Lapiz() & marcas).CuentaMarcas();
                    }
                }
                if(retval) { // Tenemos un grupo que aporta información
                    sprintf(m, "Grupo[%d] cas: %s marcas: %s", nElementos, grupo.Casillas(nElementos), grupo.Valores(marcas, nElementos));
                    msgs.push(m);
                    for(int i=0; i<nElementos; i++) grupo.Comb(i)->Amarillo();
                    for(int c=0; c<9; c++) {
                        for(int v=1; v<=9; v++) {
                            if(marcas.ContieneMarca(v) && !grupo.Contiene(fila[f].Cas(c), nElementos) && fila[f].Cas(c)->Lapiz().ContieneMarca(v)) {
                                fila[f].Cas(c)->Lapiz().EliminaMarca(v);
                                fila[f].Cas(c)->Resaltar();
                                sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                                msgs.push(m);
                            }
                        }
                    }
                    return true;
                }
            }
        } while(grupo.SiguienteCombinacion());
    }
    return false;
}

bool Tablero::ProcesaGrupoColumna(int c, int nElementos) {
    Oculto grupo(columna[c].GetCasillas());
    MarcasLapiz marcas;
    bool retval = false;

    if(grupo.PrimeraCombinacion(nElementos)) {
        do {
            marcas = MarcasLapiz();
            for(int i = 0; i<nElementos; i++) {
                if(grupo.Comb(i)->Valor()) marcas.MarcarTodas(); // Ignorar grupos con casillas finales
                else marcas = marcas | grupo.Comb(i)->Lapiz();
            }
            if(marcas.CuentaMarcas() == nElementos) {
                for(int i = 0; i<9; i++) {
                    if(!grupo.Contiene(columna[c].Cas(i), nElementos)) {
                        retval |= (columna[c].Cas(i)->Lapiz() & marcas).CuentaMarcas();
                    }
                }
                if(retval) { // Tenemos un grupo que aporta información
                    sprintf(m, "Grupo[%d] cas: %s marcas: %s", nElementos, grupo.Casillas(nElementos), grupo.Valores(marcas, nElementos));
                    msgs.push(m);
                    for(int i=0; i<nElementos; i++) grupo.Comb(i)->Amarillo();
                    for(int f=0; f<9; f++) {
                        for(int v=1; v<=9; v++) {
                            if(marcas.ContieneMarca(v) && !grupo.Contiene(columna[c].Cas(f), nElementos) && columna[c].Cas(f)->Lapiz().ContieneMarca(v)) {
                                columna[c].Cas(f)->Lapiz().EliminaMarca(v);
                                columna[c].Cas(f)->Resaltar();
                                sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                                msgs.push(m);
                            }
                        }
                    }
                    return true;
                }
            }
        } while(grupo.SiguienteCombinacion());
    }
    return false;
}

bool Tablero::ProcesaGrupoCaja(int k, int nElementos) {
    Oculto grupo(caja[k].GetCasillas());
    MarcasLapiz marcas;
    bool retval = false;

    if(grupo.PrimeraCombinacion(nElementos)) {
        do {
            marcas = MarcasLapiz();
            for(int i = 0; i<nElementos; i++) {
                if(grupo.Comb(i)->Valor()) marcas.MarcarTodas(); // Ignorar grupos con casillas finales
                else marcas = marcas | grupo.Comb(i)->Lapiz();
            }
            if(marcas.CuentaMarcas() == nElementos) {
                for(int i = 0; i<9; i++) {
                    if(!grupo.Contiene(caja[k].Cas(i), nElementos)) {
                        retval |= (caja[k].Cas(i)->Lapiz() & marcas).CuentaMarcas();
                    }
                }
                if(retval) { // Tenemos un grupo que aporta información
                    sprintf(m, "Grupo[%d] cas: %s marcas: %s", nElementos, grupo.Casillas(nElementos), grupo.Valores(marcas, nElementos));
                    msgs.push(m);
                    for(int i=0; i<nElementos; i++) grupo.Comb(i)->Amarillo();
                    for(int p=0; p<9; p++) {
                        for(int v=1; v<=9; v++) {
                            if(marcas.ContieneMarca(v) && !grupo.Contiene(caja[k].Cas(p), nElementos) && caja[k].Cas(p)->Lapiz().ContieneMarca(v)) {
                                caja[k].Cas(p)->Lapiz().EliminaMarca(v);
                                caja[k].Cas(p)->Resaltar();
                                sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                                msgs.push(m);
                            }
                        }
                    }
                    return true;
                }
            }
        } while(grupo.SiguienteCombinacion());
    }
    return false;
}

bool Tablero::ProcesaGrupo(int nElementos) {
    for(int f=0; f<9; f++) {
        if(ProcesaGrupoFila(f, nElementos)) return true;
    }
    for(int c=0; c<9; c++) {
        if(ProcesaGrupoColumna(c, nElementos)) return true;
    }
    for(int k=0; k<9; k++) {
        if(ProcesaGrupoCaja(k, nElementos)) return true;
    }
    return false;
}

bool Tablero::ProcesaXYWing() {
    int f, c;
    Casilla *pivote;
    Casilla *ala1;
    Casilla *ala2;
    int c1, f1;

    for(f=0; f<9; f++) {
        for(c=0; c<9; c++) {
            if(casilla[f][c].Lapiz().CuentaMarcas() == 2) { // Posible pivote
                pivote = &casilla[f][c];
                c1 = 0;
                do {
                    ala1 = fila[f].BuscaAla1MenosCaja(pivote, c1);
                    if(ala1) { // Tenemos posible ala 1 en f,c2 Casos 1/2
                        c1 = ala1->Columna()+1; // Probar si hay más candidatos a ala1
                        // Buscar ala2
                        ala2 = columna[c].BuscaAla2MenosCaja(pivote, ala1);
                        if(ala2) { // Patrón XY-Wing caso 1
                            if(ProcesarXYWingCaso1(pivote, ala1, ala2)) return true;
                        } else { // Posible caso 2
                            ala2 = caja[pivote->Caja()].BuscaAla2MenosFila(pivote, ala1);
                            if(ala2) { // Patrón XY-Wing caso 2
                                if(ProcesarXYWingCaso2(pivote, ala1, ala2)) return true;
                            }
                        }
                    } else { // Posible caso 3
                        c1 = 9; // No hemos encontrado el ala1 en la fila f. No seguir buscando.
                        f1 = 0;
                        do {
                            ala1 = columna[c].BuscaAla1MenosCaja(pivote, f1);
                            if(ala1) { // Tenemos posible ala 1 en f2,c
                                f1 = ala1->Fila()+1;
                                ala2 = caja[pivote->Caja()].BuscaAla2MenosColumna(pivote, ala1);
                                if(ala2) { // Patrón XY-Wing caso 3
                                    if(ProcesarXYWingCaso3(pivote, ala1, ala2)) return true;
                                }
                            }
                        } while(ala1 && f1 < 9); // Probar si hay más candidatos a ala1
                    }
                } while(ala1 && c1 < 9);
            }
        }
    }
    return false;
}

bool Tablero::ProcesarXYWingCaso1(Casilla *pivote, Casilla *ala1, Casilla *ala2) {
    MarcasLapiz comun = ala1->Lapiz() & ala2->Lapiz();
    int valor = comun.Valor();

    if(casilla[ala2->Fila()][ala1->Columna()].Lapiz().ContieneMarca(valor)) {
        sprintf(m, "XY-Wing caso 1 %c%c -> %c%c %c%c [%c]",
            'A'+pivote->Fila(), '1'+pivote->Columna(),
            'A'+ala1->Fila(), '1'+ala1->Columna(),
            'A'+ala2->Fila(), '1'+ala2->Columna(), '0'+valor);
        msgs.push(m);
        pivote->Amarillo();
        ala1->Naranja();
        ala2->Naranja();
        if(casilla[ala2->Fila()][ala1->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+ala2->Fila(), '1'+ala1->Columna());
            msgs.push(m);
            casilla[ala2->Fila()][ala1->Columna()].Resaltar();
        }
        return true;
    }
    return false;
}

bool Tablero::ProcesarXYWingCaso2(Casilla *pivote, Casilla *ala1, Casilla *ala2) {
    MarcasLapiz comun = ala1->Lapiz() & ala2->Lapiz();
    int valor = comun.Valor();
    int c1, c2, c3, c4, c5, c6;
    int k = pivote->Caja();
    int k2 = ala1->Caja();

    c1 = caja[k].Cas(0)->Columna();
    c2 = caja[k].Cas(1)->Columna();
    c3 = caja[k].Cas(2)->Columna();
    c4 = caja[k2].Cas(0)->Columna();
    c5 = caja[k2].Cas(1)->Columna();
    c6 = caja[k2].Cas(2)->Columna();
    if(casilla[ala2->Fila()][c4].Lapiz().ContieneMarca(valor) ||
       casilla[ala2->Fila()][c5].Lapiz().ContieneMarca(valor) ||
       casilla[ala2->Fila()][c6].Lapiz().ContieneMarca(valor) ||
       casilla[pivote->Fila()][c1].Lapiz().ContieneMarca(valor) ||
       casilla[pivote->Fila()][c2].Lapiz().ContieneMarca(valor) ||
       casilla[pivote->Fila()][c3].Lapiz().ContieneMarca(valor)) {
        sprintf(m, "XY-Wing caso 2 %c%c -> %c%c %c%c [%c]",
            'A'+pivote->Fila(), '1'+pivote->Columna(),
            'A'+ala1->Fila(), '1'+ala1->Columna(),
            'A'+ala2->Fila(), '1'+ala2->Columna(), '0'+valor);
        msgs.push(m);
        pivote->Amarillo();
        ala1->Naranja();
        ala2->Naranja();
        if(casilla[ala2->Fila()][c4].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+ala2->Fila(), '1'+c4);
            msgs.push(m);
            casilla[ala2->Fila()][c4].Resaltar();
        }
        if(casilla[ala2->Fila()][c5].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+ala2->Fila(), '1'+c5);
            msgs.push(m);
            casilla[ala2->Fila()][c5].Resaltar();
        }
        if(casilla[ala2->Fila()][c6].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+ala2->Fila(), '1'+c6);
            msgs.push(m);
            casilla[ala2->Fila()][c6].Resaltar();
        }
        if(casilla[pivote->Fila()][c1].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+pivote->Fila(), '1'+c1);
            msgs.push(m);
            casilla[pivote->Fila()][c1].Resaltar();
        }
        if(casilla[pivote->Fila()][c2].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+pivote->Fila(), '1'+c2);
            msgs.push(m);
            casilla[pivote->Fila()][c2].Resaltar();
        }
        if(casilla[pivote->Fila()][c3].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+pivote->Fila(), '1'+c3);
            msgs.push(m);
            casilla[pivote->Fila()][c3].Resaltar();
        }
        return true;
    }
    return false;
}

bool Tablero::ProcesarXYWingCaso3(Casilla *pivote, Casilla *ala1, Casilla *ala2) {
    MarcasLapiz comun = ala1->Lapiz() & ala2->Lapiz();
    int valor = comun.Valor();
    int f1, f2, f3, f4, f5, f6;
    int k = pivote->Caja();
    int k2 = ala1->Caja();

    f1 = caja[k].Cas(0)->Fila();
    f2 = caja[k].Cas(3)->Fila();
    f3 = caja[k].Cas(6)->Fila();
    f4 = caja[k2].Cas(0)->Fila();
    f5 = caja[k2].Cas(3)->Fila();
    f6 = caja[k2].Cas(6)->Fila();
    if(casilla[f4][ala2->Columna()].Lapiz().ContieneMarca(valor) ||
       casilla[f5][ala2->Columna()].Lapiz().ContieneMarca(valor) ||
       casilla[f6][ala2->Columna()].Lapiz().ContieneMarca(valor) ||
       casilla[f1][pivote->Columna()].Lapiz().ContieneMarca(valor) ||
       casilla[f2][pivote->Columna()].Lapiz().ContieneMarca(valor) ||
       casilla[f3][pivote->Columna()].Lapiz().ContieneMarca(valor)) {
        sprintf(m, "XY-Wing caso 3 %c%c -> %c%c %c%c [%c]",
            'A'+pivote->Fila(), '1'+pivote->Columna(),
            'A'+ala1->Fila(), '1'+ala1->Columna(),
            'A'+ala2->Fila(), '1'+ala2->Columna(), '0'+valor);
        msgs.push(m);
        pivote->Amarillo();
        ala1->Naranja();
        ala2->Naranja();
        if(casilla[f4][ala2->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f4, '1'+ala2->Columna());
            msgs.push(m);
            casilla[f4][ala2->Columna()].Resaltar();
        }
        if(casilla[f5][ala2->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f5, '1'+ala2->Columna());
            msgs.push(m);
            casilla[f5][ala2->Columna()].Resaltar();
        }
        if(casilla[f6][ala2->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f6, '1'+ala2->Columna());
            msgs.push(m);
            casilla[f6][ala2->Columna()].Resaltar();
        }
        if(casilla[f1][pivote->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f1, '1'+pivote->Columna());
            msgs.push(m);
            casilla[f1][pivote->Columna()].Resaltar();
        }
        if(casilla[f2][pivote->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f2, '1'+pivote->Columna());
            msgs.push(m);
            casilla[f1][pivote->Columna()].Resaltar();
        }
        if(casilla[f3][pivote->Columna()].Lapiz().EliminaMarca(valor)) {
            sprintf(m, "Eliminar candidato %c en %c%c", '0'+valor, 'A'+f3, '1'+pivote->Columna());
            msgs.push(m);
            casilla[f1][pivote->Columna()].Resaltar();
        }
        return true;
    }
    return false;
}

bool Tablero::ProcesaXWing() {
    Pareja par1, par2;

    for(int v=1; v<=9; v++) { // Para cada valor
        // Empezamos con las filas. No buscaremos en la última,
        // ya que no podemos comparar con las siguientes
        for(int f=0; f<8; f++) {
            // Detectar fila con dos casillas con el candidato v
            par1 = fila[f].DobleCandidato(v);
            // Si par1(0) != NULL, tenemos un candidato a X-Wing, buscaremos en el resto de filas
            if(par1.Cas(0) != NULL) {
                for(int f2=f+1; f2<9; f2++) {
                    par2 = fila[f2].DobleCandidato(v);
                    if(par2.Cas(0) != NULL &&
                       par1.Cas(0)->Columna() == par2.Cas(0)->Columna() &&
                       par1.Cas(1)->Columna() == par2.Cas(1)->Columna()) {
                        if(ProcesaXWingFila(par1, par2, v)) return true;
                    }
                }
            }
        }
        // Y ahora las columnas.
        for(int c=0; c<8; c++) {
            // Detectar fila con dos casillas con el candidato v
            par1 = columna[c].DobleCandidato(v);
            // Si par1(0) != NULL, tenemos un candidato a X-Wing, buscaremos en el resto de columnas
            if(par1.Cas(0) != NULL) {
                for(int c2=c+1; c2<9; c2++) {
                    par2 = columna[c2].DobleCandidato(v);
                    if(par2.Cas(0) != NULL &&
                       par1.Cas(0)->Fila() == par2.Cas(0)->Fila() &&
                       par1.Cas(1)->Fila() == par2.Cas(1)->Fila()) {
                        if(ProcesaXWingColumna(par1, par2, v)) return true;
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::ProcesaXWingFila(Pareja a, Pareja b, int v) {
    bool retval=false;

    for(int f=0; f<9; f++) {
        if(f != a.Cas(0)->Fila() && f != b.Cas(0)->Fila()) {
            retval |= casilla[f][a.Cas(0)->Columna()].Lapiz().ContieneMarca(v);
            retval |= casilla[f][a.Cas(1)->Columna()].Lapiz().ContieneMarca(v);
        }
    }
    if(retval) {
        sprintf(m, "X-Wing %c%c %c%c %c%c %c%c valor %d",
            'A'+a.Cas(0)->Fila(), '1'+a.Cas(0)->Columna(),
            'A'+a.Cas(1)->Fila(), '1'+a.Cas(1)->Columna(),
            'A'+b.Cas(0)->Fila(), '1'+b.Cas(0)->Columna(),
            'A'+b.Cas(1)->Fila(), '1'+b.Cas(1)->Columna(), v);
        msgs.push(m);
        a.Cas(0)->Amarillo();
        a.Cas(1)->Amarillo();
        b.Cas(0)->Amarillo();
        b.Cas(1)->Amarillo();
        for(int f=0; f<9; f++) {
            if(f != a.Cas(0)->Fila() && f != b.Cas(0)->Fila()) {
                if(casilla[f][a.Cas(0)->Columna()].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+a.Cas(0)->Columna());
                    msgs.push(m);
                    casilla[f][a.Cas(0)->Columna()].Resaltar();
                }
                if(casilla[f][a.Cas(1)->Columna()].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+a.Cas(1)->Columna());
                    msgs.push(m);
                    casilla[f][a.Cas(1)->Columna()].Resaltar();
                }
            }
        }
    }

    return retval;
}

bool Tablero::ProcesaXWingColumna(Pareja a, Pareja b, int v) {
    bool retval=false;

    for(int c=0; c<9; c++) {
        if(c != a.Cas(0)->Columna() && c != b.Cas(0)->Columna()) {
            retval |= casilla[a.Cas(0)->Fila()][c].Lapiz().ContieneMarca(v);
            retval |= casilla[a.Cas(1)->Fila()][c].Lapiz().ContieneMarca(v);
        }
    }
    if(retval) {
        sprintf(m, "X-Wing %c%c %c%c %c%c %c%c valor %d",
            'A'+a.Cas(0)->Fila(), '1'+a.Cas(0)->Columna(),
            'A'+a.Cas(1)->Fila(), '1'+a.Cas(1)->Columna(),
            'A'+b.Cas(0)->Fila(), '1'+b.Cas(0)->Columna(),
            'A'+b.Cas(1)->Fila(), '1'+b.Cas(1)->Columna(), v);
        msgs.push(m);
        a.Cas(0)->Amarillo();
        a.Cas(1)->Amarillo();
        b.Cas(0)->Amarillo();
        b.Cas(1)->Amarillo();
        for(int c=0; c<9; c++) {
            if(c != a.Cas(0)->Columna() && c != b.Cas(0)->Columna()) {
                if(casilla[a.Cas(0)->Fila()][c].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+a.Cas(0)->Fila(), '1'+c);
                    msgs.push(m);
                    casilla[a.Cas(0)->Fila()][c].Resaltar();
                }
                if(casilla[a.Cas(1)->Fila()][c].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+a.Cas(1)->Fila(), '1'+c);
                    msgs.push(m);
                    casilla[a.Cas(1)->Fila()][c].Resaltar();
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaXWingAlasFilas(int v) {
    bool retval = false;
    Pareja par;

    for(int f=0; f<9; f++) {
        // Detectar fila con dos casillas con el candidato v
        par = fila[f].DobleCandidato(v);
        // Si par.Cas(0) != NULL, y la pareja no está en la misma caja
        // tenemos un candidato a X-Wing con alas, buscaremos en el resto de filas
        if(par.Cas(0) != NULL && !par.MismaCaja()) {
            for(int f2=0; f2<9; f2++) {
                // No nos interesa que la segunda fila esté en la misma fila de cajas que la primera
                if(f/3 != f2/3) {
                    Casilla *c2 = fila[f2].DobleCandidatoYMas(par, v);
                    if(c2) {
                        retval |= ProcesaXWingAlasFilas(par, c2, v);
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaXWingAlasColumnas(int v) {
    bool retval = false;
    Pareja par;

    for(int c=0; c<9; c++) {
        // Detectar columna con dos casillas con el candidato v
        par = columna[c].DobleCandidato(v);
        // Si par.Cas(0) != NULL, y la pareja no está en la misma caja
        // tenemos un candidato a X-Wing con alas, buscaremos en el resto de columnas
        if(par.Cas(0) != NULL && !par.MismaCaja()) {
            for(int c2=0; c2<9; c2++) {
                // No nos interesa que la segunda columna esté en la misma columna de cajas que la primera
                if(c/3 != c2/3) {
                    Casilla *f2 = columna[c2].DobleCandidatoYMas(par, v);
                    if(f2) {
                        retval |= ProcesaXWingAlasColumnas(par, f2, v);
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaXWingAlasFilas(Pareja par, Casilla *c2, int v) {
    bool retval = false;
    int k2 = c2->Caja();
    int p2 = c2->PosEnCaja();

    // Verificar si el candidato está presente en la caja y columna de c2, salvo en c2
    for(int p=0; p<9; p++) {
        retval |= p != p2 && caja[k2].Cas(p)->Columna() == c2->Columna() && caja[k2].Cas(p)->Lapiz().ContieneMarca(v);
    }
    if(retval) {
        sprintf(m, "X-Wing Alas %c%c %c%c %c%c %c%c Valor %d",
                'A'+par.Cas(0)->Fila(), '1'+par.Cas(0)->Columna(),
                'A'+par.Cas(1)->Fila(), '1'+par.Cas(1)->Columna(),
                'A'+c2->Fila(), '1'+par.Cas(0)->Columna(),
                'A'+c2->Fila(), '1'+par.Cas(1)->Columna(), v);
        msgs.push(m);
        par.Cas(0)->Amarillo();
        par.Cas(1)->Amarillo();
        casilla[c2->Fila()][par.Cas(0)->Columna()].Amarillo();
        casilla[c2->Fila()][par.Cas(1)->Columna()].Amarillo();
        // Eliminar el candidato si está presente en la caja y columna de c2, salvo en c2
        for(int p=0; p<9; p++) {
            if(p != p2 && caja[k2].Cas(p)->Columna() == c2->Columna() && caja[k2].Cas(p)->Lapiz().EliminaMarca(v))
                caja[k2].Cas(p)->Resaltar();
        }

        // Resaltar las alas:
        for(int p=0; p<9; p++) {
            if(p != p2 && caja[k2].Cas(p)->Fila() == c2->Fila() && caja[k2].Cas(p)->Lapiz().ContieneMarca(v))
                caja[k2].Cas(p)->Naranja();
        }
    }
    return retval;
}

bool Tablero::ProcesaXWingAlasColumnas(Pareja par, Casilla *f2, int v) {
    bool retval = false;
    int k2 = f2->Caja();
    int p2 = f2->PosEnCaja();

    // Verificar si el candidato está presente en la caja y fila de f2, salvo en f2
    for(int p=0; p<9; p++) {
        retval |= p != p2 && caja[k2].Cas(p)->Fila() == f2->Fila() && caja[k2].Cas(p)->Lapiz().ContieneMarca(v);
    }

    if(retval) {
        sprintf(m, "X-Wing Alas %c%c %c%c %c%c %c%c Valor %d",
                'A'+par.Cas(0)->Fila(), '1'+par.Cas(0)->Columna(),
                'A'+par.Cas(1)->Fila(), '1'+par.Cas(1)->Columna(),
                'A'+par.Cas(0)->Fila(), '1'+f2->Columna(),
                'A'+par.Cas(1)->Fila(), '1'+f2->Columna(), v);
        msgs.push(m);
        par.Cas(0)->Amarillo();
        par.Cas(1)->Amarillo();
        casilla[par.Cas(0)->Fila()][f2->Columna()].Amarillo();
        casilla[par.Cas(1)->Fila()][f2->Columna()].Amarillo();
        // Eliminar el candidato si está presente en la caja y columna de f2, salvo en f2
        for(int p=0; p<9; p++) {
            if(p != p2 && caja[k2].Cas(p)->Fila() == f2->Fila() && caja[k2].Cas(p)->Lapiz().EliminaMarca(v))
                caja[k2].Cas(p)->Resaltar();
        }

        // Resaltar las alas:
        for(int p=0; p<9; p++) {
            if(p != p2 && caja[k2].Cas(p)->Columna() == f2->Columna() && caja[k2].Cas(p)->Lapiz().ContieneMarca(v))
                caja[k2].Cas(p)->Naranja();
        }
    }
    return retval;
}

bool Tablero::ProcesaXWingAlas() {
    for(int v=1; v<=9; v++) {
        // Empezamos con las filas.
        if(BuscaXWingAlasFilas(v)) return true;
        // Y ahora las columnas.
        if(BuscaXWingAlasColumnas(v)) return true;
    }
    return false;
}

bool Tablero::BuscaSwordFishFilas(int v) {
    bool retval=false;
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;

    for(int f1=0; f1<7; f1++) {
        // Buscamos en las primeras 7 filas en las que no aparezca v
        // y esté como candidato en 2, 3 columnas:
        if(!fila[f1].Presente(v) && fila[f1].CuentaCandidato(v) <= 3) {
            // Ahora buscamos las filas f1+1 a septima otra que tenga 2, 3 candidatos v.
            for(int f2= f1+1; f2<8; f2++) {
                if(!fila[f2].Presente(v) && fila[f2].CuentaCandidato(v) <= 3) {
                    // Si entre las dos filas, el candidato aparece en 2, 3 columnas, buscamos otra fila
                    CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v);
                    if(CasillasColumnas.CuentaCasillas() == 3) {
                        // Ahora buscamos las filas f2+1 a novena otra que tenga 2, 3 candidatos v en las
                        // mismas columnas que las anteriores.
                        for(int f3=f2+1; f3<9; f3++) {
                            if(!fila[f3].Presente(v) && fila[f3].CuentaCandidato(v) <= 3) {
                                CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v)+fila[f3].Marcas(v);
                                if(CasillasColumnas.CuentaCasillas() == 3) {
                                    // Tenemos una medusa.
                                    CasillasFilas.Borra();
                                    CasillasFilas.AsignaCasilla(f1);
                                    CasillasFilas.AsignaCasilla(f2);
                                    CasillasFilas.AsignaCasilla(f3);
                                    if(ProcesaSwordFishFilas(CasillasFilas, CasillasColumnas, v)) return true;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaSwordFishColumnas(int v) {
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;

    for(int c1=0; c1<7; c1++) {
        // Buscamos en las primeras 6 columnas en las que no aparezca v
        // y esté como candidato en 2 o 3 columnas:
        if(!columna[c1].Presente(v) && columna[c1].CuentaCandidato(v) <= 3) {
            // Ahora buscamos las columnas c1+1 a septima otra que tenga 2, 3 o 4 candidatos v.
            for(int c2= c1+1; c2<8; c2++) {
                if(!columna[c2].Presente(v) && columna[c2].CuentaCandidato(v) <= 3) {
                    // Si entre las dos columnas, el candidato aparece en tres filas, buscamos otra columna
                    CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v);
                    if(CasillasFilas.CuentaCasillas() == 3) {
                        // Ahora buscamos las columnas c2+1 a octava otra que tenga 2, 3 o 4 candidatos v en las
                        // mismas filas que las anteriores.
                        for(int c3=c2+1; c3<9; c3++) {
                            if(!columna[c3].Presente(v) && columna[c3].CuentaCandidato(v) <= 3) {
                                CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v)+columna[c3].Marcas(v);
                                if(CasillasFilas.CuentaCasillas() == 3) {
                                    CasillasColumnas.Borra();
                                    CasillasColumnas.AsignaCasilla(c1);
                                    CasillasColumnas.AsignaCasilla(c2);
                                    CasillasColumnas.AsignaCasilla(c3);
                                    if(ProcesaSwordFishColumnas(CasillasFilas, CasillasColumnas, v)) return true;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::ProcesaSwordFishFilas(MarcasCasillas fils, MarcasCasillas cols, int v) {
    bool retval = false;
    int ff[3], i=0;
    int cc[3], j=0;

    for(int f=0; f<9; f++) {
        if(!fils.ContieneCasilla(f)) {
            for(int c=0; c<9; c++) {
                if(cols.ContieneCasilla(c)) {
                    retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                }
            }
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(fils.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(cols.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Pez espada F: fila [%c,%c,%c] col [%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], '1'+cc[0], '1'+cc[1], '1'+cc[2], v);
        msgs.push(m);
        for(i=0; i<3; i++) {
            for(j=0; j<3; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }
        for(int f=0; f<9; f++) {
            if(!fils.ContieneCasilla(f)) {
                for(int c=0; c<9; c++) {
                    if(cols.ContieneCasilla(c)) {
                        if(casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaSwordFishColumnas(MarcasCasillas fils, MarcasCasillas cols, int v) {
    bool retval = false;
    int ff[3], i=0;
    int cc[3], j=0;

    for(int c=0; c<9; c++) {
        if(!cols.ContieneCasilla(c)) {
            for(int f=0; f<9; f++) {
                if(fils.ContieneCasilla(f)) {
                    retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                }
            }
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(fils.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(cols.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Pez espada C: fila [%c,%c,%c] col [%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], '1'+cc[0], '1'+cc[1], '1'+cc[2], v);
        msgs.push(m);
        for(i=0; i<3; i++) {
            for(j=0; j<3; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }
        for(int c=0; c<9; c++) {
            if(!cols.ContieneCasilla(c)) {
                for(int f=0; f<9; f++) {
                    if(fils.ContieneCasilla(f)) {
                        if(casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaSwordFish() {
    for(int v=1; v<=9; v++) {
        if(BuscaSwordFishFilas(v)) return true;
        if(BuscaSwordFishColumnas(v)) return true;
    }
    return false;
}

bool Tablero::TodasConDosCandidatosMenosUna() {
    int cuenta3Candidatos=0;
    Casilla *cas3;

    // Buscar en cada casilla, si el número de candidatos en más de una es tres
    // o en alguna hay más de tres, retornar false
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].Lapiz().CuentaMarcas() > 3) return false;
            if(casilla[f][c].Lapiz().CuentaMarcas() == 3) { cas3 = &casilla[f][c]; cuenta3Candidatos++; }
            if(cuenta3Candidatos > 1) return false;
        }
    }
    if(cuenta3Candidatos == 0) return false; // En éste caso el sudoku estará resuelto
    // Si no hemos retornado false, tenemos un patrón en el que todas las casillas
    // con candidatos tienen dos, menos una que tiene 3.
    // La casilla (f3,c3) es la que tiene tres candidatos, en la fila f3, la columna c3 y
    // en la caja donde está esa casilla hay un candidato que aparece 3 veces, localizarlo.
    for(int v=1; v<=9; v++) {
        if(cas3->Lapiz().ContieneMarca(v) && fila[cas3->Fila()].CuentaCandidato(v) == 3) {
            sprintf(m, "Todas las casillas tienen dos candidatos menos %c%c", 'A'+cas3->Fila(), '1'+cas3->Columna());
            msgs.push(m);
            sprintf(m, "$Valor final %d en %c%c", v, 'A'+cas3->Fila(), '1'+cas3->Columna());
            msgs.push(m);
            // Marcar valor final v en cas3
            Asignar(cas3->Fila(), cas3->Columna(), v);
            cas3->Resaltar();
        }
    }
    return true;
}

bool Tablero::SolucionUnicaA() {

    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].ParDesnudo(Casilla::enfila) && c/3 == casilla[f][c].ParDesnudo(Casilla::enfila)->Columna()/3) {
                if(BuscaFila(casilla[f][c])) return true;
            } else
            if(casilla[f][c].ParDesnudo(Casilla::encolumna) && f/3 == casilla[f][c].ParDesnudo(Casilla::encolumna)->Fila()/3) {
                if(BuscaColumna(casilla[f][c])) return true;
            }
        }
    }
    return false;
}

bool Tablero::BuscaFila(Casilla& cas) {
    int c1, c2, ff;
    int nMarcas1, nMarcas2;
    Casilla *cas3 = NULL;

    ff = cas.Fila();
    c1 = cas.Columna();
    c2 = cas.ParDesnudo(Casilla::enfila)->Columna();
    for(int f=0; f<9; f++) {
        if(f != ff) {
            nMarcas1 = casilla[f][c1].Lapiz().CuentaMarcas();
            nMarcas2 = casilla[f][c2].Lapiz().CuentaMarcas();

            if(nMarcas1==2 && nMarcas2>=3 &&
               casilla[f][c1].Lapiz() == cas.Lapiz() &&
               (casilla[f][c2].Lapiz() & cas.Lapiz()) == cas.Lapiz()) {
                cas3 = &casilla[f][c2];
            } else
            if(nMarcas1>=3 && nMarcas2==2 &&
               casilla[f][c2].Lapiz() == cas.Lapiz() &&
               (casilla[f][c1].Lapiz() & cas.Lapiz()) == cas.Lapiz()) {
                cas3 = &casilla[f][c1];
            }

            if(cas3) {
                sprintf(m, "Solucion unica A %c%c %c%c %c%c %c%c valores %s",
                    'A'+ff, '1'+c1, 'A'+ff, '1'+c2, 'A'+cas3->Fila(), '1'+c1, 'A'+cas3->Fila(), '1'+c2,
                    cas.Lapiz().Valores());
                msgs.push(m);
                cas.Amarillo();
                cas.ParDesnudo(Casilla::enfila)->Amarillo();
                if(c1 != cas3->Columna()) casilla[cas3->Fila()][c1].Amarillo();
                if(c2 != cas3->Columna()) casilla[cas3->Fila()][c2].Amarillo();
                casilla[cas3->Fila()][cas3->Columna()].Resaltar();

                cas3->Lapiz().EliminaMarca(cas.Lapiz().ValorMarca(1));
                sprintf(m, "Eliminar candidato %d en %c%c", cas.Lapiz().ValorMarca(1), 'A'+cas3->Fila(), '1'+cas3->Columna());
                msgs.push(m);
                cas3->Lapiz().EliminaMarca(cas.Lapiz().ValorMarca(2));
                sprintf(m, "Eliminar candidato %d en %c%c", cas.Lapiz().ValorMarca(2), 'A'+cas3->Fila(), '1'+cas3->Columna());
                msgs.push(m);

                return true;
            }
        }
    }
    return false;
}

bool Tablero::BuscaColumna(Casilla& cas) {
    int f1, f2, cc;
    int nMarcas1, nMarcas2;
    Casilla *cas3 = NULL;

    cc = cas.Columna();
    f1 = cas.Fila();
    f2 = cas.ParDesnudo(Casilla::encolumna)->Fila();
    for(int c=0; c<9; c++) {
        if(c != cc) {
            nMarcas1 = casilla[f1][c].Lapiz().CuentaMarcas();
            nMarcas2 = casilla[f1][c].Lapiz().CuentaMarcas();

            if(nMarcas1==2 && nMarcas2>=3 &&
               casilla[f1][c].Lapiz() == cas.Lapiz() &&
               (casilla[f2][c].Lapiz() & cas.Lapiz()) == cas.Lapiz()) {
                cas3 = &casilla[f2][c];
            } else
            if(nMarcas1>=3 && nMarcas2==2 &&
               casilla[f2][c].Lapiz() == cas.Lapiz() &&
               (casilla[f1][c].Lapiz() & cas.Lapiz()) == cas.Lapiz()) {
                cas3 = &casilla[f1][c];
            }

            if(cas3) {
                sprintf(m, "Solucion unica A %c%c %c%c %c%c %c%c valores %s",
                    'A'+f1, '1'+cc, 'A'+f2, '1'+cc, 'A'+f1, '1'+cas3->Columna(), 'A'+f2, '1'+cas3->Columna(),
                    cas.Lapiz().Valores());
                msgs.push(m);
                cas.Amarillo();
                cas.ParDesnudo(Casilla::encolumna)->Amarillo();
                if(f1 != cas3->Fila()) casilla[f1][cas3->Columna()].Amarillo();
                if(f1 != cas3->Fila()) casilla[f2][cas3->Columna()].Amarillo();
                casilla[cas3->Fila()][cas3->Columna()].Resaltar();

                cas3->Lapiz().EliminaMarca(cas.Lapiz().ValorMarca(1));
                sprintf(m, "Eliminar candidato %d en %c%c", cas.Lapiz().ValorMarca(1), 'A'+cas3->Fila(), '1'+cas3->Columna());
                msgs.push(m);
                cas3->Lapiz().EliminaMarca(cas.Lapiz().ValorMarca(2));
                sprintf(m, "Eliminar candidato %d en %c%c", cas.Lapiz().ValorMarca(2), 'A'+cas3->Fila(), '1'+cas3->Columna());
                msgs.push(m);

                return true;
            }
        }
    }
    return false;
}


bool Tablero::SolucionUnicaB() {
    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].ParDesnudo(Casilla::enfila) && c/3 == casilla[f][c].ParDesnudo(Casilla::enfila)->Columna()/3) {
                if(BuscaFilaTrios(casilla[f][c])) return true;
            }
            if(casilla[f][c].ParDesnudo(Casilla::encolumna) && f/3 == casilla[f][c].ParDesnudo(Casilla::encolumna)->Fila()/3) {
                if(BuscaColumnaTrios(casilla[f][c])) return true;
            }
            // Variantes B'
            if(casilla[f][c].ParDesnudo(Casilla::enfila) && c/3 != casilla[f][c].ParDesnudo(Casilla::enfila)->Columna()/3) {
                if(BuscaFilaTriosB(casilla[f][c])) return true;
            }
            if(casilla[f][c].ParDesnudo(Casilla::encolumna) && f/3 != casilla[f][c].ParDesnudo(Casilla::encolumna)->Fila()/3) {
                if(BuscaColumnaTriosB(casilla[f][c])) return true;
            }
        }
    }

    return false;
}

bool Tablero::BuscaFilaTrios(Casilla& cas) {
    bool retval=false;
    int c1, c2, ff;
    MarcasLapiz marcas;
    int v;

    ff = cas.Fila();
    c1 = cas.Columna();
    c2 = cas.ParDesnudo(Casilla::enfila)->Columna();
    for(int f=0; f<9; f++) {
        if(f != ff) {
            if(casilla[f][c1].Lapiz().CuentaMarcas()==3 && casilla[f][c2].Lapiz().CuentaMarcas()==3 &&
               casilla[f][c1].Lapiz() == casilla[f][c2].Lapiz() &&
               (casilla[f][c1].Lapiz() & cas.Lapiz()) == (casilla[f][c2].Lapiz() & cas.Lapiz())) {
                marcas = casilla[f][c1].Lapiz() - cas.Lapiz(); // Marca extra
                v = marcas.Valor(); // Valor de la marca extra
                // Eliminar v del resto de la fila f
                for(int c=0; c<9; c++)
                    if(c != c1 && c!= c2) retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                int k = casilla[f][c1].Caja();
                for(int p=0; p<9; p++)
                    if(caja[k].Cas(p)->Fila() != f) retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);

                if(retval) {
                    sprintf(m, "Solucion unica B %c%c %c%c %c%c %c%c valor %d",
                        'A'+ff, '1'+c1, 'A'+ff, '1'+c2, 'A'+f, '1'+c1, 'A'+f, '1'+c2, v);
                    msgs.push(m);
                    cas.Amarillo();
                    cas.ParDesnudo(Casilla::enfila)->Amarillo();
                    casilla[f][c1].Amarillo();
                    casilla[f][c2].Amarillo();
                    for(int c=0; c<9; c++) {
                        if(c != c1 && c!= c2 && casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                    int k = casilla[f][c1].Caja();
                    for(int p=0; p<9; p++) {
                        if(caja[k].Cas(p)->Fila() != f && caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                            msgs.push(m);
                            caja[k].Cas(p)->Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaColumnaTrios(Casilla& cas) {
    bool retval=false;
    int f1, f2, cc;
    MarcasLapiz marcas;
    int v;

    cc = cas.Columna();
    f1 = cas.Fila();
    f2 = cas.ParDesnudo(Casilla::encolumna)->Fila();
    for(int c=0; c<9; c++) {
        if(c != cc) {
            if(casilla[f1][c].Lapiz().CuentaMarcas()==3 && casilla[f2][c].Lapiz().CuentaMarcas()==3 &&
               casilla[f1][c].Lapiz() == casilla[f2][c].Lapiz() &&
               (casilla[f1][c].Lapiz() & cas.Lapiz()) == (casilla[f2][c].Lapiz() & cas.Lapiz())) {
                marcas = casilla[f1][c].Lapiz() - cas.Lapiz(); // Marca extra
                v = marcas.Valor(); // Valor de la marca extra
                // Eliminar v del resto de la columna c
                for(int f=0; f<9; f++)
                    if(f != f1 && f!= f2) retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                int k = casilla[f1][c].Caja();
                for(int p=0; p<9; p++)
                    if(caja[k].Cas(p)->Columna() != c) retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);

                if(retval) {
                    sprintf(m, "Solucion unica B %c%c %c%c %c%c %c%c valor %d",
                        'A'+f1, '1'+cc, 'A'+f2, '1'+cc, 'A'+f1, '1'+c, 'A'+f2, '1'+c, v);
                    msgs.push(m);
                    cas.Amarillo();
                    cas.ParDesnudo(Casilla::encolumna)->Amarillo();
                    casilla[f1][c].Amarillo();
                    casilla[f2][c].Amarillo();
                    for(int f=0; f<9; f++) {
                        if(f != f1 && f!= f2 && casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                    int k = casilla[f1][c].Caja();
                    for(int p=0; p<9; p++) {
                        if(caja[k].Cas(p)->Columna() != c && caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                            msgs.push(m);
                            caja[k].Cas(p)->Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaFilaTriosB(Casilla& cas) {
    bool retval=false;
    int c1, c2, ff;
    MarcasLapiz marcas;
    int v;

    ff = cas.Fila();
    c1 = cas.Columna();
    c2 = cas.ParDesnudo(Casilla::enfila)->Columna();
    for(int f=0; f<9; f++) {
        if(f != ff && f/3==ff/3) {
            if(casilla[f][c1].Lapiz().CuentaMarcas()==3 && casilla[f][c2].Lapiz().CuentaMarcas()==3 &&
               casilla[f][c1].Lapiz() == casilla[f][c2].Lapiz() &&
               (casilla[f][c1].Lapiz() & cas.Lapiz()) == (casilla[f][c2].Lapiz() & cas.Lapiz())) {
                marcas = casilla[f][c1].Lapiz() - cas.Lapiz(); // Marca extra
                v = marcas.Valor(); // Valor de la marca extra
                // Eliminar v del resto de la fila f
                for(int c=0; c<9; c++)
                    if(c != c1 && c!= c2) retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                if(retval) {
                    sprintf(m, "Solucion unica B' %c%c %c%c %c%c %c%c valor %d",
                        'A'+ff, '1'+c1, 'A'+ff, '1'+c2, 'A'+f, '1'+c1, 'A'+f, '1'+c2, v);
                    msgs.push(m);
                    cas.Amarillo();
                    cas.ParDesnudo(Casilla::enfila)->Amarillo();
                    casilla[f][c1].Amarillo();
                    casilla[f][c2].Amarillo();
                    for(int c=0; c<9; c++) {
                        if(c != c1 && c!= c2 && casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaColumnaTriosB(Casilla& cas) {
    bool retval=false;
    int f1, f2, cc;
    MarcasLapiz marcas;
    int v;

    cc = cas.Columna();
    f1 = cas.Fila();
    f2 = cas.ParDesnudo(Casilla::encolumna)->Fila();
    for(int c=0; c<9; c++) {
        if(c != cc && c/3==cc/3) {
            if(casilla[f1][c].Lapiz().CuentaMarcas()==3 && casilla[f2][c].Lapiz().CuentaMarcas()==3 &&
               casilla[f1][c].Lapiz() == casilla[f2][c].Lapiz() &&
               (casilla[f1][c].Lapiz() & cas.Lapiz()) == (casilla[f2][c].Lapiz() & cas.Lapiz())) {
                marcas = casilla[f1][c].Lapiz() - cas.Lapiz(); // Marca extra
                v = marcas.Valor(); // Valor de la marca extra
                // Eliminar v del resto de la columna c
                for(int f=0; f<9; f++)
                    if(f != f1 && f!= f2) retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                if(retval) {
                    sprintf(m, "Solucion unica B %c%c %c%c %c%c %c%c valor %d",
                        'A'+f1, '1'+cc, 'A'+f2, '1'+cc, 'A'+f1, '1'+c, 'A'+f2, '1'+c, v);
                    msgs.push(m);
                    cas.Amarillo();
                    cas.ParDesnudo(Casilla::encolumna)->Amarillo();
                    casilla[f1][c].Amarillo();
                    casilla[f2][c].Amarillo();
                    for(int f=0; f<9; f++) {
                        if(f != f1 && f!= f2 && casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

// Para cada casilla con tres candidatos buscamos dos alas compatibles con un patrón XYZ-Wing
bool Tablero::ProcesaXYZWing() {
    MarcasLapiz m[3];
    int i, j, k, v;
    int pos[3][3]; // array con zonas, indice 1 (f, c, k) e indice 2 con las marcas,

    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(casilla[f][c].Lapiz().CuentaMarcas() == 3) {
                // Calcular las tres combinaciones de marcas de lápiz
                // empezamos con las tres, y eliminamos una distinta en cada una
                m[0] = m[1] = m[2] = casilla[f][c].Lapiz();
                i=0;
                for(int v=1; v<=9; v++) {
                    if(m[i].ContieneMarca(v)) { m[i++].EliminaMarca(v); }
                }
                k = casilla[f][c].Caja();
                // Buscar en misma fila, columna y caja casillas con dos candidatos coincidentes
                for(i = 0; i<3; i++) {
                    pos[0][i] = fila[f].BuscarMarcas(m[i]);
                    pos[1][i] = columna[c].BuscarMarcas(m[i]);
                    pos[2][i] = caja[k].BuscarMarcas(m[i]);
                }
                // Solo puede haber un patron XYZ-Wing si se ha detectado una casilla con marcas en la caja
                for(i=0; i<3; i++) {
                    if(pos[2][i] != -1) { // Tenemos el par m[i] en la pos[2][i] de la caja
                        for(j=0; j<3; j++) { // Buscar las otras dos parejas en fila o columna
                            if(pos[0][j] != -1 && i!=j) { // Tenemos el par m[j] en la columna pos[0][j]
                                v = (m[i] & m[j]).Valor(); // Valor común de m[i] y m[j]
                                if(f!=caja[k].Cas(pos[2][i])->Fila() && casilla[f][pos[0][j]].Caja() != k)
                                   if(ProcesarXYZFila(k, f, c, v, pos[0][j], pos[2][i])) return true;
                            }
                            if(pos[1][j] != -1 && i!=j) { // Tenemos el par m[j] en la fila pos[1][j]
                                v = (m[i] & m[j]).Valor(); // Valor común de m[i] y m[j]
                                if(c != caja[k].Cas(pos[2][i])->Columna() && casilla[pos[1][j]][c].Caja() != k)
                                    if(ProcesarXYZColumna(k, f, c, v, pos[1][j], pos[2][i])) return true;
                            }
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::ProcesarXYZFila(int k, int fil, int col, int v, int c, int p) {
    bool retval=false;

    // Hay que eliminar el candidato v en la fila fil de la caja k,
    // excepto en la columna en que está el pivote.
    for(int p=0; p<9; p++) {
        if(caja[k].Cas(p)->Fila() == fil && caja[k].Cas(p)->Columna() != col)
            retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);
    }
    if(retval) {
        sprintf(m, "XYZ-Wing en %c%c -> %c%c %c%c", 'A'+fil, '1'+col,
            'A'+fil, '1'+c, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
        msgs.push(m);
        casilla[fil][col].Naranja();
        casilla[fil][c].Amarillo();
        caja[k].Cas(p)->Amarillo();
        for(int p=0; p<9; p++) {
            if(caja[k].Cas(p)->Fila() == fil && caja[k].Cas(p)->Columna() != col)
                if(caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                    msgs.push(m);
                    caja[k].Cas(p)->Resaltar();
                }
        }
    }
    return retval;
}

bool Tablero::ProcesarXYZColumna(int k, int fil, int col, int v, int f, int p) {
    bool retval=false;

    // Hay que eliminar el candidato v en la columna col de la caja k,
    // excepto en la fila en que está el pivote.
    for(int p=0; p<9; p++) {
        if(caja[k].Cas(p)->Columna() == col && caja[k].Cas(p)->Fila() != fil)
            retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);
    }
    if(retval) {
        sprintf(m, "XYZ-Wing en %c%c -> %c%c %c%c", 'A'+fil, '1'+col,
            'A'+f, '1'+col, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
        msgs.push(m);
        casilla[fil][col].Naranja();
        casilla[f][col].Amarillo();
        caja[k].Cas(p)->Amarillo();
        for(int p=0; p<9; p++) {
            if(caja[k].Cas(p)->Columna() == col && caja[k].Cas(p)->Fila() != fil)
                if(caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                    msgs.push(m);
                    caja[k].Cas(p)->Resaltar();
                }
        }
    }
    return retval;
}

bool Tablero::ProcesaSashimi() {
    if(BuscaSashimiFilas()) return true;
    if(BuscaSashimiColumnas()) return true;
    return false;
}

bool Tablero::BuscaSashimiFilas() {
    Pareja par;
    int c1, c2;
    int c1a, c1b, c2a, c2b;

    for(int v=1; v<=9; v++) {
        for(int f1=0; f1<8; f1++) {
            par = fila[f1].DobleCandidato(v);
            if(par.Cas(0) && !par.MismaCaja() ) {
                c1 = par.Cas(0)->Columna();
                c2 = par.Cas(1)->Columna();
                // Calcular las columnas donde pueden estar las alas:
                switch(c1%3) {
                    case 0: c1a=3*(c1/3)+1; c1b=3*(c1/3)+2; break;
                    case 1: c1a=3*(c1/3)+0; c1b=3*(c1/3)+2; break;
                    case 2: c1a=3*(c1/3)+0; c1b=3*(c1/3)+1; break;
                }
                switch(c2%3) {
                    case 0: c2a=3*(c2/3)+1; c2b=3*(c2/3)+2; break;
                    case 1: c2a=3*(c2/3)+0; c2b=3*(c2/3)+2; break;
                    case 2: c2a=3*(c2/3)+0; c2b=3*(c2/3)+1; break;
                }
                for(int f2=0; f2<9; f2++) {
                    // Las dos filas deben estar en cajas diferentes, y la segunda debe tener el candidato tres veces
                    if(casilla[f1][c1].Caja() != casilla[f2][c1].Caja() && fila[f2].CuentaCandidato(v) == 3) {
                        if(casilla[f2][c1].Lapiz().ContieneMarca(v) && !casilla[f2][c2].Lapiz().ContieneMarca(v) &&
                           casilla[f2][c2a].Lapiz().ContieneMarca(v) && casilla[f2][c2b].Lapiz().ContieneMarca(v)){
                            if(ProcesarSashimiFila(f1, f2, c1, c2, v, c2, c2a, c2b)) return true;
                        }
                        if(casilla[f2][c2].Lapiz().ContieneMarca(v) && !casilla[f2][c1].Lapiz().ContieneMarca(v) &&
                           casilla[f2][c1a].Lapiz().ContieneMarca(v) && casilla[f2][c1b].Lapiz().ContieneMarca(v)) {
                            if(ProcesarSashimiFila(f1, f2, c1, c2, v, c1, c1a, c1b)) return true;
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::BuscaSashimiColumnas() {
    Pareja par;
    int f1, f2;
    int f1a, f1b, f2a, f2b;

    for(int v=1; v<=9; v++) {
        for(int c1=0; c1<8; c1++) {
            par = columna[c1].DobleCandidato(v);
            if(par.Cas(0) && !par.MismaCaja()) {
                f1 = par.Cas(0)->Fila();
                f2 = par.Cas(1)->Fila();
                // Calcular las filas donde pueden estar las alas:
                switch(f1%3) {
                    case 0: f1a=3*(f1/3)+1; f1b=3*(f1/3)+2; break;
                    case 1: f1a=3*(f1/3)+0; f1b=3*(f1/3)+2; break;
                    case 2: f1a=3*(f1/3)+0; f1b=3*(f1/3)+1; break;
                }
                switch(f2%3) {
                    case 0: f2a=3*(f2/3)+1; f2b=3*(f2/3)+2; break;
                    case 1: f2a=3*(f2/3)+0; f2b=3*(f2/3)+2; break;
                    case 2: f2a=3*(f2/3)+0; f2b=3*(f2/3)+1; break;
                }
                for(int c2=0; c2<9; c2++) {
                    // Las dos columnas deben estar en cajas diferentes, y la segunda debe tener el candidato tres veces
                    if(casilla[f1][c1].Caja() != casilla[f1][c2].Caja() && columna[c2].CuentaCandidato(v) == 3) {
                        if(casilla[f1][c2].Lapiz().ContieneMarca(v) && !casilla[f2][c2].Lapiz().ContieneMarca(v) &&
                           casilla[f2a][c2].Lapiz().ContieneMarca(v) && casilla[f2b][c2].Lapiz().ContieneMarca(v)){
                            if(ProcesarSashimiColumna(f1, f2, c1, c2, v, f2, f2a, f2b)) return true;
                        }
                        if(casilla[f2][c2].Lapiz().ContieneMarca(v) && !casilla[f1][c2].Lapiz().ContieneMarca(v) &&
                           casilla[f1a][c2].Lapiz().ContieneMarca(v) && casilla[f1b][c2].Lapiz().ContieneMarca(v)) {
                            if(ProcesarSashimiColumna(f1, f2, c1, c2, v, f1, f1a, f1b)) return true;
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::ProcesarSashimiFila(int f1, int f2, int c1, int c2, int v, int c, int ca, int cb) {
    int retval = false;
    int k = casilla[f2][c].Caja();

    for(int p=0; p<9; p++) {
        if(caja[k].Cas(p)->Columna() == c) retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);
    }
    if(retval) {
        if(c==c2)
            sprintf(m, "Sashimi X-Wing F %c%c %c%c %c%c %c%c %c%c, valor %d",
                'A'+f1, '1'+c1, 'A'+f2, '1'+c1, 'A'+f1, '1'+c2, 'A'+f2, '1'+ca, 'A'+f2, '1'+cb, v);
        else
            sprintf(m, "Sashimi X-Wing F %c%c %c%c %c%c %c%c %c%c, valor %d",
                'A'+f1, '1'+c1, 'A'+f2, '1'+ca, 'A'+f2, '1'+cb, 'A'+f1, '1'+c2, 'A'+f2, '1'+c2, v);
        msgs.push(m);
        casilla[f1][c1].Amarillo();
        casilla[f1][c2].Amarillo();
        casilla[f2][ca].Naranja();
        casilla[f2][cb].Naranja();
        if(c==c2)
            casilla[f2][c1].Amarillo();
        else
            casilla[f2][c2].Amarillo();
        for(int p=0; p<9; p++) {
            if(caja[k].Cas(p)->Columna() == c)
                if(caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                    msgs.push(m);
                    caja[k].Cas(p)->Resaltar();
                }
        }
    }
    return retval;
}

bool Tablero::ProcesarSashimiColumna(int f1, int f2, int c1, int c2, int v, int f, int fa, int fb) {
    int retval = false;
    int k = casilla[f][c2].Caja();

    for(int p=0; p<9; p++) {
        if(caja[k].Cas(p)->Fila() == f) retval |= caja[k].Cas(p)->Lapiz().ContieneMarca(v);
    }
    if(retval) {
        if(f==f2)
            sprintf(m, "Sashimi X-Wing C %c%c %c%c %c%c %c%c %c%c, valor %d",
                'A'+f1, '1'+c1, 'A'+f1, '1'+c2, 'A'+f2, '1'+c1, 'A'+fa, '1'+c2, 'A'+fb, '1'+c2, v);
        else
            sprintf(m, "Sashimi X-Wing C %c%c %c%c %c%c %c%c %c%c, valor %d",
                'A'+f1, '1'+c1, 'A'+f1, '1'+c2, 'A'+fa, '1'+c2, 'A'+fb, '1'+c2, 'A'+f2, '1'+c2, v);
        msgs.push(m);
        casilla[f1][c1].Amarillo();
        casilla[f2][c1].Amarillo();
        casilla[fa][c2].Naranja();
        casilla[fb][c2].Naranja();
        if(f==f2)
            casilla[f1][c2].Amarillo();
        else
            casilla[f2][c2].Amarillo();
        for(int p=0; p<9; p++) {
            if(caja[k].Cas(p)->Fila() == f)
                if(caja[k].Cas(p)->Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+caja[k].Cas(p)->Fila(), '1'+caja[k].Cas(p)->Columna());
                    msgs.push(m);
                    caja[k].Cas(p)->Resaltar();
                }
        }
    }
    return retval;
}

bool Tablero::ProcesaSkyscraper() {
    if(BuscaSkyscraperFilas()) return true;
    if(BuscaSkyscraperColumnas()) return true;
    return false;
}

bool Tablero::BuscaSkyscraperFilas() {
    Pareja par1, par2;

    for(int v=1; v<=9; v++) {
        for(int f1=0; f1<8; f1++) {
            par1 = fila[f1].DobleCandidato(v);
            // La pareja de candidatos no puede estar en la misma caja
            if(par1.Cas(0) && par1.Cas(0)->Caja() != par1.Cas(1)->Caja()) {
                for(int f2=f1+1; f2<9; f2++) {
                    par2 = fila[f2].DobleCandidato(v);
                    if(par2.Cas(0)) {
                        // Las casillas de cada pareja deben estar en las misma columna de cajas, y en distintas cajas
                        if(par1.Cas(0)->Caja()%3 == par2.Cas(0)->Caja()%3 && par1.Cas(1)->Caja()%3 == par2.Cas(1)->Caja()%3 &&
                            par1.Cas(0)->Caja() != par2.Cas(0)->Caja()){
                            // Segunda condición:
                            // Una de las casillas de cada pareja debe estar en la misma columna
                            if(par1.Cas(0)->Columna() == par2.Cas(0)->Columna() && par1.Cas(1)->Columna() != par2.Cas(1)->Columna()) {
                                if(ProcesarSkyscraperFila(par1, par2, 1, v)) return true;
                            }
                            if(par1.Cas(0)->Columna() != par2.Cas(0)->Columna() && par1.Cas(1)->Columna() == par2.Cas(1)->Columna()) {
                                if(ProcesarSkyscraperFila(par1, par2, 0, v)) return true;
                            }
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::BuscaSkyscraperColumnas() {
    Pareja par1, par2;

    for(int v=1; v<=9; v++) {
        for(int c1=0; c1<8; c1++) {
            par1 = columna[c1].DobleCandidato(v);
            // La pareja de candidatos no puede estar en la misma caja
            if(par1.Cas(0) && par1.Cas(0)->Caja() != par1.Cas(1)->Caja()) {
                for(int c2=c1+1; c2<9; c2++) {
                    par2 = columna[c2].DobleCandidato(v);
                    if(par2.Cas(0)) {
                        // Las casillas de cada pareja deben estar en las misma fila de cajas, y en distintas cajas
                        if(par1.Cas(0)->Caja()/3 == par2.Cas(0)->Caja()/3 && par1.Cas(1)->Caja()/3 == par2.Cas(1)->Caja()/3 &&
                            par1.Cas(0)->Caja() != par2.Cas(0)->Caja()){
                            // Segunda condición:
                            // Una de las casillas de cada pareja debe estar en la misma columna
                            if(par1.Cas(0)->Fila() == par2.Cas(0)->Fila() && par1.Cas(1)->Fila() != par2.Cas(1)->Fila()) {
                                if(ProcesarSkyscraperColumna(par1, par2, 1, v)) return true;
                            }
                            if(par1.Cas(0)->Fila() != par2.Cas(0)->Fila() && par1.Cas(1)->Fila() == par2.Cas(1)->Fila()) {
                                if(ProcesarSkyscraperColumna(par1, par2, 0, v)) return true;
                            }
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::ProcesarSkyscraperFila(Pareja &par1, Pareja &par2, int np, int v) {
    bool retval = false;
    int k1, k2;

    k1 = par1.Cas(np)->Caja();
    k2 = par2.Cas(np)->Caja();
    for(int f=0; f<9; f++) {
        if(casilla[f][par1.Cas(np)->Columna()].Caja() == k2) {
            retval |= casilla[f][par1.Cas(np)->Columna()].Lapiz().ContieneMarca(v);
        }
        if(casilla[f][par2.Cas(np)->Columna()].Caja() == k1) {
            retval |= casilla[f][par2.Cas(np)->Columna()].Lapiz().ContieneMarca(v);
        }
    }
    if(retval) {
        sprintf(m, "Skyscraper F %c%c %c%c %c%c %c%c, %d",
            'A'+par1.Cas(0)->Fila(), '1'+par1.Cas(0)->Columna(), 'A'+par1.Cas(1)->Fila(), '1'+par1.Cas(1)->Columna(),
            'A'+par2.Cas(0)->Fila(), '1'+par2.Cas(0)->Columna(), 'A'+par2.Cas(1)->Fila(), '1'+par2.Cas(1)->Columna(), v);
        msgs.push(m);
        casilla[par1.Cas(0)->Fila()][par1.Cas(0)->Columna()].Amarillo();
        casilla[par1.Cas(1)->Fila()][par1.Cas(1)->Columna()].Amarillo();
        casilla[par2.Cas(0)->Fila()][par2.Cas(0)->Columna()].Amarillo();
        casilla[par2.Cas(1)->Fila()][par2.Cas(1)->Columna()].Amarillo();
        for(int f=0; f<9; f++) {
            if(casilla[f][par1.Cas(np)->Columna()].Caja() == k2) {
                if(casilla[f][par1.Cas(np)->Columna()].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %s", v, casilla[f][par1.Cas(np)->Columna()].Nombre());
                    msgs.push(m);
                    casilla[f][par1.Cas(np)->Columna()].Resaltar();
                }
            }
            if(casilla[f][par2.Cas(np)->Columna()].Caja() == k1) {
                if(casilla[f][par2.Cas(np)->Columna()].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %s", v, casilla[f][par2.Cas(np)->Columna()].Nombre());
                    msgs.push(m);
                    casilla[f][par2.Cas(np)->Columna()].Resaltar();
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesarSkyscraperColumna(Pareja &par1, Pareja &par2, int np, int v) {
    bool retval = false;
    int k1, k2;

    k1 = par1.Cas(np)->Caja();
    k2 = par2.Cas(np)->Caja();
    for(int c=0; c<9; c++) {
        if(casilla[par1.Cas(np)->Fila()][c].Caja() == k2) {
            retval |= casilla[par1.Cas(np)->Fila()][c].Lapiz().ContieneMarca(v);
        }
        if(casilla[par2.Cas(np)->Fila()][c].Caja() == k1) {
            retval |= casilla[par2.Cas(np)->Fila()][c].Lapiz().ContieneMarca(v);
        }
    }
    if(retval) {
        sprintf(m, "Skyscraper C %c%c %c%c %c%c %c%c, %d",
            'A'+par1.Cas(0)->Fila(), '1'+par1.Cas(0)->Columna(), 'A'+par1.Cas(1)->Fila(), '1'+par1.Cas(1)->Columna(),
            'A'+par2.Cas(0)->Fila(), '1'+par2.Cas(0)->Columna(), 'A'+par2.Cas(1)->Fila(), '1'+par2.Cas(1)->Columna(), v);
        msgs.push(m);
        casilla[par1.Cas(0)->Fila()][par1.Cas(0)->Columna()].Amarillo();
        casilla[par1.Cas(1)->Fila()][par1.Cas(1)->Columna()].Amarillo();
        casilla[par2.Cas(0)->Fila()][par2.Cas(0)->Columna()].Amarillo();
        casilla[par2.Cas(1)->Fila()][par2.Cas(1)->Columna()].Amarillo();
        for(int c=0; c<9; c++) {
            if(casilla[par1.Cas(np)->Fila()][c].Caja() == k2) {
                if(casilla[par1.Cas(np)->Fila()][c].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %s", v, casilla[par1.Cas(np)->Fila()][c].Nombre());
                    msgs.push(m);
                    casilla[par1.Cas(np)->Fila()][c].Resaltar();
                }
            }
            if(casilla[par2.Cas(np)->Fila()][c].Caja() == k1) {
                if(casilla[par2.Cas(np)->Fila()][c].Lapiz().EliminaMarca(v)) {
                    sprintf(m, "Eliminar candidato %d en %s", v, casilla[par2.Cas(np)->Fila()][c].Nombre());
                    msgs.push(m);
                    casilla[par2.Cas(np)->Fila()][c].Resaltar();
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaRectangulosVacios() {
    int rv;
    int f, c;
    int f2, c2;
    Pareja par;

    for(int v=1; v<=9; v++) {
        // En cada caja buscar patrones de rectángulos vacíos
        for(int k=0; k<9; k++) {
            rv = caja[k].RectanguloVacio(v);
            if(-1 != rv) {
                f = 3*(k/3) + rv/3;
                c = 3*(k%3) + rv%3;
                // Buscar par puntero en columnas con un candiadto en fila f
                for(int c1=0; c1<9; c1++) {
                    par = columna[c1].DobleCandidato(v);
                    if(par.Cas(0) && // Si encuentra un par
                       !par.MismaCaja() && // El par no está en la misma caja
                       (par.Cas(0)->Caja() != k) && (par.Cas(1)->Caja() != k) &&// Y está en una fila de cajas distinta a la del rectángulo
                       (par.Cas(0)->Fila() == f || par.Cas(1)->Fila() == f)) { // Y una de las casillas coincide con la fila en el rectángulo
                        if(par.Cas(0)->Fila() == f) f2 = par.Cas(1)->Fila();
                        else f2 = par.Cas(0)->Fila();
                        if(casilla[f2][c].Lapiz().ContieneMarca(v)) {
                            sprintf(m, "Rectangulo vacio en caja %d, valor %d", 1+k, v);
                            msgs.push(m);
                            sprintf(m, "Si %d en columna %c en caja %d, no puede estar en %c%c", v, '1'+c, 1+k, 'A'+f2, '1'+c);
                            msgs.push(m);
                            sprintf(m, "Si %d en fila %c en caja %d, no puede estar en %c%c, debe estar en %c%c y no puede estar (%c%c)",
                                v, 'A'+f, 1+k, 'A'+f, '1'+c1, 'A'+f2, '1'+c1, 'A'+f2, '1'+c);
                            msgs.push(m);
                            for(int p=0; p<9; p++) if(caja[k].Cas(p)->Lapiz().ContieneMarca(v)) caja[k].Cas(p)->Amarillo();
                            par.Cas(0)->Naranja();
                            par.Cas(1)->Naranja();
                            casilla[f2][c].Lapiz().EliminaMarca(v);
                            casilla[f2][c].Resaltar();
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f2, '1'+c);
                            msgs.push(m);
                            return true;
                        }
                    }
                }
                // Buscar par puntero en filas con un candiadto en fila c
                for(int f1=0; f1<9; f1++) {
                    par = fila[f1].DobleCandidato(v);
                    if(par.Cas(0) && !par.MismaCaja() &&
                       (par.Cas(0)->Caja() != k) && (par.Cas(1)->Caja() != k) &&
                       (par.Cas(0)->Columna() == c || par.Cas(1)->Columna() == c)) {
                        if(par.Cas(0)->Columna() == c) c2 = par.Cas(1)->Columna();
                        else c2 = par.Cas(0)->Columna();
                        if(casilla[f][c2].Lapiz().ContieneMarca(v)) {
                            sprintf(m, "Rectangulo vacio en caja %d, valor %d", 1+k, v);
                            msgs.push(m);
                            sprintf(m, "Si %d en fila %c en caja %d, no puede estar en %c%c", v, 'A'+f, 1+k, 'A'+f, '1'+c2);
                            msgs.push(m);
                            sprintf(m, "Si %d en columna %c en caja %d, no puede estar en %c%c, debe estar en %c%c y no puede estar (%c%c)",
                                v, '1'+c, 1+k, 'A'+f1, '1'+c, 'A'+f1, '1'+c2, 'A'+f, '1'+c2);
                            msgs.push(m);
                            for(int p=0; p<9; p++) if(caja[k].Cas(p)->Lapiz().ContieneMarca(v)) caja[k].Cas(p)->Amarillo();
                            par.Cas(0)->Naranja();
                            par.Cas(1)->Naranja();
                            casilla[f][c2].Lapiz().EliminaMarca(v);
                            casilla[f][c2].Resaltar();
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c2);
                            msgs.push(m);
                            return true;
                        }
                    }
                }
            }
        }
    }
    return false;
}

bool Tablero::BuscaJellyFishFilas(int v) {
    bool retval=false;
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;

    for(int f1=0; f1<6; f1++) {
        // Buscamos en las primeras 6 filas en las que no aparezca v
        // y esté como candidato en 2, 3 o 4 columnas:
        if(!fila[f1].Presente(v) && fila[f1].CuentaCandidato(v) <= 4) {
            // Ahora buscamos las filas f1+1 a septima otra que tenga 2, 3 o 4 candidatos v.
            for(int f2= f1+1; f2<7; f2++) {
                if(!fila[f2].Presente(v) && fila[f2].CuentaCandidato(v) <= 4) {
                    // Si entre las dos filas, el candidato aparece en 3 o 4 columnas, buscamos otra fila
                    CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v);
                    if(CasillasColumnas.CuentaCasillas() <= 4) {
                        // Ahora buscamos las filas f2+1 a octava otra que tenga 2, 3 o 4 candidatos v en las
                        // mismas columnas que las anteriores.
                        for(int f3=f2+1; f3<8; f3++) {
                            if(!fila[f3].Presente(v) && fila[f3].CuentaCandidato(v) <= 4) {
                                CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v)+fila[f3].Marcas(v);
                                if(CasillasColumnas.CuentaCasillas() <= 4) {
                                    // Ahora buscamos las filas f3+1 a novena otra que tenga 2, 3 o 4 candidatos v en las
                                    // mismas columnas que las anteriores.
                                    for(int f4=f3+1; f4<9; f4++) {
                                        if(!fila[f4].Presente(v) && fila[f4].CuentaCandidato(v) <= 4) {
                                            CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v)+fila[f3].Marcas(v)+fila[f4].Marcas(v);
                                            if(CasillasColumnas.CuentaCasillas() == 4) {
                                                // Tenemos una medusa.
                                                CasillasFilas.AsignaCasilla(f1);
                                                CasillasFilas.AsignaCasilla(f2);
                                                CasillasFilas.AsignaCasilla(f3);
                                                CasillasFilas.AsignaCasilla(f4);
                                                if(ProcesaJellyFishFilas(CasillasFilas, CasillasColumnas, v)) return true;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaJellyFishColumnas(int v) {
    bool retval=false;
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;

    for(int c1=0; c1<6; c1++) {
        // Buscamos en las primeras 6 columnas en las que no aparezca v
        // y esté como candidato en 2 o 3 columnas:
        if(!columna[c1].Presente(v) && columna[c1].CuentaCandidato(v) <= 4) {
            // Ahora buscamos las columnas c1+1 a septima otra que tenga 2, 3 o 4 candidatos v.
            for(int c2= c1+1; c2<7; c2++) {
                if(!columna[c2].Presente(v) && columna[c2].CuentaCandidato(v) <= 4) {
                    // Si entre las dos columnas, el candidato aparece en tres filas, buscamos otra columna
                    CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v);
                    if(CasillasFilas.CuentaCasillas() <= 4) {
                        // Ahora buscamos las columnas c2+1 a octava otra que tenga 2, 3 o 4 candidatos v en las
                        // mismas filas que las anteriores.
                        for(int c3=c2+1; c3<8; c3++) {
                            if(!columna[c3].Presente(v) && columna[c3].CuentaCandidato(v) <= 4) {
                                CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v)+columna[c3].Marcas(v);
                                if(CasillasFilas.CuentaCasillas() <= 4) {
                                    // Ahora buscamos las columnas c3+1 a novena otra que tenga 2, 3 o 4 candidatos v en las
                                    // mismas columnas que las anteriores.
                                    for(int c4=c3+1; c4<9; c4++) {
                                        if(!columna[c4].Presente(v) && columna[c4].CuentaCandidato(v) <= 4) {
                                            CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v)+columna[c3].Marcas(v)+columna[c4].Marcas(v);
                                            if(CasillasFilas.CuentaCasillas() == 4) {
                                                // Tenemos una medusa.
                                                CasillasColumnas.AsignaCasilla(c1);
                                                CasillasColumnas.AsignaCasilla(c2);
                                                CasillasColumnas.AsignaCasilla(c3);
                                                CasillasColumnas.AsignaCasilla(c4);
                                                if(ProcesaJellyFishColumnas(CasillasFilas, CasillasColumnas, v)) return true;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaJellyFishFilas(MarcasCasillas fils, MarcasCasillas cols, int v) {
    bool retval = false;
    int ff[4], i=0;
    int cc[4], j=0;

    for(int f=0; f<9; f++) {
        if(!fils.ContieneCasilla(f)) {
            for(int c=0; c<9; c++) {
                if(cols.ContieneCasilla(c)) {
                    retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                }
            }
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(fils.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(cols.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Medusa F: fila [%c,%c,%c,%c] col [%c,%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], 'A'+ff[3], '1'+cc[0], '1'+cc[1], '1'+cc[2], '1'+cc[3], v);
        msgs.push(m);
        for(i=0; i<4; i++) {
            for(j=0; j<4; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }
        for(int f=0; f<9; f++) {
            if(!fils.ContieneCasilla(f)) {
                for(int c=0; c<9; c++) {
                    if(cols.ContieneCasilla(c)) {
                        if(casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaJellyFishColumnas(MarcasCasillas fils, MarcasCasillas cols, int v) {
    bool retval = false;
    int ff[4], i=0;
    int cc[4], j=0;

    for(int c=0; c<9; c++) {
        if(!cols.ContieneCasilla(c)) {
            for(int f=0; f<9; f++) {
                if(fils.ContieneCasilla(f)) {
                    retval |= casilla[f][c].Lapiz().ContieneMarca(v);
                }
            }
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(fils.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(cols.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Medusa C: fila [%c,%c,%c,%c] col [%c,%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], 'A'+ff[3], '1'+cc[0], '1'+cc[1], '1'+cc[2], '1'+cc[3], v);
        msgs.push(m);
        for(i=0; i<4; i++) {
            for(j=0; j<4; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }
        for(int c=0; c<9; c++) {
            if(!cols.ContieneCasilla(c)) {
                for(int f=0; f<9; f++) {
                    if(fils.ContieneCasilla(f)) {
                        if(casilla[f][c].Lapiz().EliminaMarca(v)) {
                            sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
                            msgs.push(m);
                            casilla[f][c].Resaltar();
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaJellyFish() {
    bool retval=false;

    for(int v=1; v<=9; v++) {
        retval |= BuscaJellyFishFilas(v);
        retval |= BuscaJellyFishColumnas(v);
    }
    return retval;
}

bool Tablero::StringKite() {
    Pareja par;
    Pareja parF, parC;

    for(int v=1; v<=9; v++) {
        for(int k=0; k<9; k++) {
            par = caja[k].DobleCandidato(v);
            if(par.Cas(0) && !par.MismaFila() && !par.MismaColumna()) {
                // Buscar par en columna par.Cas(0) y en fila par.Cas(1)
                parC = columna[par.Cas(0)->Columna()].DobleCandidato(v);
                parF = fila[par.Cas(1)->Fila()].DobleCandidato(v);
                if(parC.Cas(0) && parF.Cas(0) && ProcesarStringKite(par, parF, parC, v)) return true;

                // Buscar par en fila par.Cas(0) y en columna par.Cas(1)
                parC = columna[par.Cas(1)->Columna()].DobleCandidato(v);
                parF = fila[par.Cas(0)->Fila()].DobleCandidato(v);
                if(parC.Cas(0) && parF.Cas(0) && ProcesarStringKite(par, parF, parC, v)) return true;
            }
        }
    }
    return false;
}

bool Tablero::ProcesarStringKite(Pareja &par, Pareja &parF, Pareja &parC, int v) {
    int f, c;

    if(par.Cas(0) == parC.Cas(0) || par.Cas(1) == parC.Cas(0))
        f = parC.Cas(1)->Fila();
    else
        f = parC.Cas(0)->Fila();
    if(par.Cas(0) == parF.Cas(0) || par.Cas(1) == parF.Cas(0))
        c = parF.Cas(1)->Columna();
    else
        c = parF.Cas(0)->Columna();
    if(casilla[f][c].Lapiz().ContieneMarca(v)) {
        sprintf(m, "SringKite caja %c valor %d", '1'+par.Cas(0)->Caja(), v);
        msgs.push(m);
        parC.Cas(0)->Amarillo();
        parC.Cas(1)->Amarillo();
        parF.Cas(0)->Amarillo();
        parF.Cas(1)->Amarillo();
        sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c);
        msgs.push(m);
        casilla[f][c].Resaltar();
        casilla[f][c].Lapiz().EliminaMarca(v);
        return true;
    }
    return false;
}

bool Tablero::SwordFishAlas() {
    bool retval=false;

    for(int v=1; v<=9; v++) {
        retval |= BuscaSwordFishAlasFilas(v);
        retval |= BuscaSwordFishAlasColumnas(v);
    }
    return retval;
}

bool Tablero::BuscaSwordFishAlasFilas(int v) {
    bool retval=false;
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;
    MarcasCasillas CasillasAlas;
    int caj;

    for(int f1=0; f1<8; f1++) {
        // Buscamos en las primeras 8 filas en las que no aparezca v
        // y esté como candidato en 2, 3 columnas:
        if(!fila[f1].Presente(v) && fila[f1].CuentaCandidato(v) <= 3) {
            // Ahora buscamos las filas f1+1 a novena otra que tenga 2, 3 candidatos v.
            for(int f2= f1+1; f2<9; f2++) {
                if(!fila[f2].Presente(v) && fila[f2].CuentaCandidato(v) <= 3) {
                    // Si entre las dos filas, el candidato aparece en 2, 3 columnas, buscamos otra fila
                    // que complete un patrón pez espada con alas:
                    CasillasColumnas = fila[f1].Marcas(v)+fila[f2].Marcas(v);
                    if(CasillasColumnas.CuentaCasillas() == 3) {
                        // Ahora buscamos en todas las filas menos f1 y f2 otra que tenga al menos 3 candidatos v
                        for(int f3=0; f3<9; f3++) {
                            if(f3 != f1 && f3 != f2 && !fila[f3].Presente(v) &&
                               fila[f3].CuentaCandidato(v) >= 3 &&
                               (fila[f3].Marcas(v) & CasillasColumnas).CuentaCasillas() >= 2) {
                                CasillasAlas = fila[f3].Marcas(v) & ~CasillasColumnas;
                                caj = CasillasAlas.Cajas();
                                if(((caj & CasillasColumnas.Cajas()) == caj) && (caj == 1 || caj == 2 || caj == 4)) {
                                    // CasillasColumnas tiene las columnas de casillas con el candidato
                                    CasillasFilas.Borra();
                                    CasillasFilas.AsignaCasilla(f1);
                                    CasillasFilas.AsignaCasilla(f2);
                                    CasillasFilas.AsignaCasilla(f3);
                                    if(ProcesaSwordFishAlasFilas(CasillasFilas, CasillasColumnas, CasillasAlas, f3, v)) return true;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::BuscaSwordFishAlasColumnas(int v) {
    bool retval=false;
    MarcasCasillas CasillasFilas;
    MarcasCasillas CasillasColumnas;
    MarcasCasillas CasillasAlas;
    int caj;

    for(int c1=0; c1<8; c1++) {
        // Buscamos en las primeras 8 columnas en las que no aparezca v
        // y esté como candidato en 2, 3 filas:
        if(!columna[c1].Presente(v) && columna[c1].CuentaCandidato(v) <= 3) {
            // Ahora buscamos las columnas c1+1 a novena otra que tenga 2, 3 candidatos v.
            for(int c2= c1+1; c2<9; c2++) {
                if(!columna[c2].Presente(v) && columna[c2].CuentaCandidato(v) <= 3) {
                    // Si entre las dos columnas, el candidato aparece en 2, 3 filas, buscamos otra columna
                    // que complete un patrón pez espada con alas:
                    CasillasFilas = columna[c1].Marcas(v)+columna[c2].Marcas(v);
                    if(CasillasFilas.CuentaCasillas() == 3) {
                        // Ahora buscamos en todas las Columnas menos c1 y c2 otra que tenga al menos 3 candidatos v
                        for(int c3=0; c3<9; c3++) {
                            if(c3 != c1 && c3 != c2 && !columna[c3].Presente(v) &&
                               columna[c3].CuentaCandidato(v) >= 3 &&
                               (columna[c3].Marcas(v) & CasillasFilas).CuentaCasillas() >= 2) {
                                CasillasAlas = columna[c3].Marcas(v) & ~CasillasFilas;
                                caj = CasillasAlas.Cajas();
                                if(((caj & CasillasFilas.Cajas()) == caj) && (caj == 1 || caj == 2 || caj == 4)) {
                                    // CasillasFilas tiene las filas de casillas con el candidato
                                    CasillasColumnas.Borra();
                                    CasillasColumnas.AsignaCasilla(c1);
                                    CasillasColumnas.AsignaCasilla(c2);
                                    CasillasColumnas.AsignaCasilla(c3);
                                    if(ProcesaSwordFishAlasColumnas(CasillasFilas, CasillasColumnas, CasillasAlas, c3, v)) return true;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaSwordFishAlasFilas(MarcasCasillas &CasillasFilas,
                                        MarcasCasillas &CasillasColumnas,
                                        MarcasCasillas &CasillasAlas, int f3, int v) {
    bool retval = false;
    int ca, c2;
    MarcasCasillas ck;
    int caj;
    int ff[3], cc[3], i=0, j=0;

    caj = CasillasAlas.Cajas();
    ca = CasillasAlas.Casilla();
    ck = CasillasColumnas;
    ck.MascaraCaja(caj);
    c2 = ck.Casilla();
    // Eliminar candidatos en columna de candidato que comparte caja con [f3][c]
    // menos en filas f1, f2 y f3
    for(int f=0; f<9; f++) {
        if(!CasillasFilas.ContieneCasilla(f) && casilla[f][ca].Caja() == casilla[f3][c2].Caja()) {
            retval |= casilla[f][c2].Lapiz().ContieneMarca(v);
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(CasillasFilas.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(CasillasColumnas.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Pez espada Alas F: fila [%c,%c,%c] col [%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], '1'+cc[0], '1'+cc[1], '1'+cc[2], v);
        msgs.push(m);
        casilla[f3][ca].Naranja();
        for(i=0; i<3; i++) {
            for(j=0; j<3; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }

        for(int f=0; f<9; f++) {
            if(!CasillasFilas.ContieneCasilla(f) && casilla[f][ca].Caja() == casilla[f3][c2].Caja()) {
                if(casilla[f][c2].Lapiz().EliminaMarca(v)) {
                    casilla[f][c2].Resaltar();
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f, '1'+c2);
                    msgs.push(m);
                }
            }
        }
    }
    return retval;
}

bool Tablero::ProcesaSwordFishAlasColumnas(MarcasCasillas &CasillasFilas,
                                           MarcasCasillas &CasillasColumnas,
                                           MarcasCasillas &CasillasAlas, int c3, int v) {
    bool retval = false;
    int fa, f2;
    MarcasCasillas fk;
    int caj;
    int ff[3], cc[3], i=0, j=0;

    caj = CasillasAlas.Cajas();
    fa = CasillasAlas.Casilla();
    fk = CasillasFilas;
    fk.MascaraCaja(caj);
    f2 = fk.Casilla();
    // Eliminar candidatos en fila de candidato que comparte caja con [f][c3]
    // menos en columnas c1, c2 y c3
    for(int c=0; c<9; c++) {
        if(!CasillasColumnas.ContieneCasilla(c) && casilla[fa][c].Caja() == casilla[f2][c3].Caja()) {
            retval |= casilla[f2][c].Lapiz().ContieneMarca(v);
        }
    }

    if(retval) {
        for(int f=0; f<9; f++) if(CasillasFilas.ContieneCasilla(f)) ff[i++] = f;
        for(int c=0; c<9; c++) if(CasillasColumnas.ContieneCasilla(c)) cc[j++] = c;
        sprintf(m, "Pez espada Alas C: fila [%c,%c,%c] col [%c,%c,%c] valor %d",
                'A'+ff[0], 'A'+ff[1], 'A'+ff[2], '1'+cc[0], '1'+cc[1], '1'+cc[2], v);
        msgs.push(m);
        casilla[fa][c3].Naranja();
        for(i=0; i<3; i++) {
            for(j=0; j<3; j++) {
                if(casilla[ff[i]][cc[j]].Lapiz().ContieneMarca(v))
                    casilla[ff[i]][cc[j]].Amarillo();
            }
        }

        for(int c=0; c<9; c++) {
            if(!CasillasColumnas.ContieneCasilla(c) && casilla[fa][c].Caja() == casilla[f2][c3].Caja()) {
                if(casilla[f2][c].Lapiz().EliminaMarca(v)) {
                    casilla[f2][c].Resaltar();
                    sprintf(m, "Eliminar candidato %d en %c%c", v, 'A'+f2, '1'+c);
                    msgs.push(m);
                }
            }
        }
    }
    return retval;
}

bool Tablero::VerificarTerminadoCorrecto() {
    MarcasLapiz Marcas;

    for(int f=0; f<9; f++) {
        Marcas = MarcasLapiz();
        for(int c=0; c<9; c++) {
            Marcas.AsignaMarca(fila[f].Cas(c)->Valor());
        }
        if(Marcas.CuentaMarcas()<9) return false;
    }

    for(int c=0; c<9; c++) {
        Marcas = MarcasLapiz();
        for(int f=0; f<9; f++) {
            Marcas.AsignaMarca(columna[c].Cas(f)->Valor());
        }
        if(Marcas.CuentaMarcas()<9) return false;
    }

    for(int k=0; k<9; k++) {
        Marcas = MarcasLapiz();
        for(int p=0; p<9; p++) {
            Marcas.AsignaMarca(caja[k].Cas(p)->Valor());
        }
        if(Marcas.CuentaMarcas()<9) return false;
    }
    return true;
}

bool Tablero::VerificarSolucionPosible() {

    for(int f=0; f<9; f++) {
        for(int c=0; c<9; c++) {
            if(!casilla[f][c].Valor() && casilla[f][c].Lapiz().CuentaMarcas()==0) return false;
        }
    }
    return true;
}
